/*******************************************************************************

   MBS270 V2 solar power

   Copyright Mobisense Systems 2009

   This is a sample program monitoring a solar power source for the board.
   The solar power is made of a solar panel and Li-Ion battery. It is mounted
   on a pan tilt platform to optimize the panel position vs sun.
   The solar panel voltage is monitored via an ADC channel. The pan tilt
   platform is made with two RC servos.
   MBS270 can be directly powered with the Li-Ion battery via the MBSPOW-4V
   power module and the servos with the 5V output making a fully autonomous
   power supply.

   The application handles the various situations with a state machine.
   After initialization, the program looks for the best orientation and tests
   the panel voltage output. If it is higher than the sunlight threshold, the
   program goes in sunlight mode. If it is lower than the darkness threshold,
   it goes in darkness mode. Otherwise it goes in lowlight mode.
   In sunlight mode, the panel does not move for sl_tempo (which is relatively
   short to optimize the charge).
   In lowlight mode, the panel does not move for ll_tempo (which is relatively
   long to save power).
   In darkness mode, the panel is oriented towards sunrise for dn_tempo.
   Once the corresponding tempo has ellapsed, panel voltage is tested as
   described above. If it is higher than the darkness threshold, the program
   looks for the best orientation (tracking state) and determines the next
   state.

   Pan reference:
    South direction in north hemisphere, east is positive, west is negative.
    North direction in south hemisphere, west is positive, east is negative.
   Tilt reference: horizontal plane, upwards is +90°.

*******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/ucb1400_adc.h>
#include <linux/pxa_servo.h>

#define DEBUG
/* #define DEBUG2 */
#define ADC_DEV_NAME		"/dev/ucb1400-adc"
#define SERVO_DEV_NAME		"/dev/pxa-servo"
#define ADC_MAX_INPUT		7630	/* max input voltage in mV */
#define SEARCH_STEP		5
#define SERVO_SPEED		45	/* deg/second */

/* default settings */
#define DEFAULT_ADC_CHANNEL	0
#define DEFAULT_PAN_GPIO	16
#define DEFAULT_TILT_GPIO	115
#define DEFAULT_HEMISPHERE	1
#define DEFAULT_SL_THR		3700
#define DEFAULT_DN_THR		2000
#define DEFAULT_SL_TEMPO	600
#define DEFAULT_LL_TEMPO	3600
#define DEFAULT_DN_TEMPO	7200
#define DEFAULT_PAN_MIN		-90
#define DEFAULT_PAN_MAX		90
#define DEFAULT_PAN		DEFAULT_PAN_MIN
#define DEFAULT_DN_PAN		90
#define DEFAULT_TILT_MIN	0
#define DEFAULT_TILT_MAX	90
#define DEFAULT_TILT		45
#define DEFAULT_DN_TILT		0
#define DEFAULT_SERVO_GAIN	11

/* Possible states */
typedef enum {
	MSP_STATE_SUNLIGHT,
	MSP_STATE_LOWLIGHT,
	MSP_STATE_DARKNESS,
	MSP_STATE_TRACKING
} msp_state_t;

typedef struct msp_servo {
	int gpio;
	unsigned int id;
	int gain; /* in microsecond/degree */
	int angle, default_pos, min, max, darkness_pos; /* in degree */
} msp_servo_t;

typedef struct msp_context {
	msp_state_t state;
	int servo_fd, adc_fd;
	int adc_channel;
	int panel_voltage; /* in mV */
	int hemisphere; /* 1 = North, -1 = South */
	int sl_thr, dn_thr; /* voltage thresholds in mV */
	int sl_tempo, ll_tempo, dn_tempo; /* tempo in seconds for each state */
	msp_servo_t pan, tilt;
} msp_context_t;


void init_defaults(msp_context_t *ctx)
{
	ctx->state = -1;
	ctx->servo_fd = -1;
	ctx->adc_fd = -1;
	ctx->adc_channel = DEFAULT_ADC_CHANNEL;
	ctx->panel_voltage = 0;
	ctx->hemisphere = DEFAULT_HEMISPHERE;
	ctx->sl_thr = DEFAULT_SL_THR;
	ctx->dn_thr = DEFAULT_DN_THR;
	ctx->sl_tempo = DEFAULT_SL_TEMPO;
	ctx->ll_tempo = DEFAULT_LL_TEMPO;
	ctx->dn_tempo = DEFAULT_DN_TEMPO;
	ctx->pan.gpio = DEFAULT_PAN_GPIO;
	ctx->pan.id = 0xffffffff;
	ctx->pan.gain = -DEFAULT_SERVO_GAIN;
	ctx->pan.angle = 0;
	ctx->pan.default_pos = DEFAULT_PAN;
	ctx->pan.min = DEFAULT_PAN_MIN;
	ctx->pan.max = DEFAULT_PAN_MAX;
	ctx->pan.darkness_pos = DEFAULT_DN_PAN;
	ctx->tilt.gpio = DEFAULT_TILT_GPIO;
	ctx->tilt.id = 0xffffffff;
	ctx->tilt.gain = -DEFAULT_SERVO_GAIN;
	ctx->tilt.angle = 0;
	ctx->tilt.default_pos = DEFAULT_TILT;
	ctx->tilt.min = DEFAULT_TILT_MIN;
	ctx->tilt.max = DEFAULT_TILT_MAX;
	ctx->tilt.darkness_pos = DEFAULT_DN_TILT;
}

void display_context(msp_context_t *ctx, int full, char *str)
{
	time_t current_time;
	time(&current_time);
	printf("\n");
	if (str != NULL) printf("%s - ", str);
	printf("solar_power context - %s", ctime(&current_time));
	printf("State = %d\n", ctx->state);
	if (full) {
		printf("File descriptors: servo = %d, ADC = %d\n",
		       ctx->servo_fd, ctx->adc_fd);
		printf("Hemisphere = ");
		if (ctx->hemisphere == 1) printf("north");
		else printf("south");
		printf("\n");
		printf("ADC channel = %d - Panel voltage = %dmV\n",
		       ctx->adc_channel, ctx->panel_voltage);
		printf("pan GPIO = %d  - tilt GPIO = %d\n",
		       ctx->pan.gpio, ctx->tilt.gpio);
		printf("Solar panel thresholds: sunlight = %dmV, darkness = %dmV\n",
		       ctx->sl_thr, ctx->dn_thr);
		printf("Tempos: sunlight = %ds, lowlight = %ds, darkness = %ds\n",
		       ctx->sl_tempo, ctx->ll_tempo, ctx->dn_tempo);
		printf("Servo IDs: pan = %d, tilt = %d\n",
		       ctx->pan.id, ctx->tilt.id);
		printf("Servo gain in microseconds/deg: pan = %d, tilt = %d\n",
		       ctx->pan.gain, ctx->tilt.gain);
		printf("Pan : current = %ddeg, default = %ddeg, min = %ddeg, max = %ddeg, \
darkness = %ddeg\n",
		       ctx->pan.angle, ctx->pan.default_pos,
		       ctx->pan.min, ctx->pan.max, ctx->pan.darkness_pos);
		printf("Tilt: current = %ddeg, default = %ddeg, min = %ddeg, max = %ddeg, \
darkness = %ddeg\n",
		       ctx->tilt.angle, ctx->tilt.default_pos,
		       ctx->tilt.min, ctx->tilt.max, ctx->tilt.darkness_pos);
	}
}

int read_args(msp_context_t *ctx, int argc, char *argv[])
{
	int aa, ii, nb_param, *dest[3];
	
	aa = 1;
	while (aa < argc) {
		if (strcmp(argv[aa], "-hemisphere") == 0) {
			if (aa+1 >= argc) goto call_error;
			if (strcmp(argv[aa+1], "north") == 0)
				ctx->hemisphere = 1;
			else if (strcmp(argv[aa+1], "south") == 0)
				ctx->hemisphere = -1;
			else {
				printf("\n\nERROR: -hemisphere option parameter\
 must be north or south.\n\n");
				goto call_error;
			}
			aa+=2;
		}
		else if (strcmp(argv[aa], "-adc_channel") == 0) {
			if (aa+1 >= argc) goto call_error;
			int channel;
			channel = atoi(argv[aa+1]);
			if ((channel < 0) || (channel > 3)) {
				printf("\n\nERROR: ADC channel must be between\
 0 and 3.\n\n");
				goto call_error;
			}
			else
				ctx->adc_channel = channel;
			aa+=2;
		}
		else if (strcmp(argv[aa], "-thresholds") == 0) {
			int thr;
			nb_param = 2;
			if (aa+nb_param >= argc) goto call_error;
			dest[0] = &ctx->sl_thr;
			dest[1] = &ctx->dn_thr;
			for (ii=0; ii<nb_param; ii++) {
				thr = atoi(argv[aa+1+ii]);
				if ((thr < 0) || (thr > ADC_MAX_INPUT)) {
					printf("\n\nERROR: voltage thresholds\
 must be between 0 and %d\n\n", ADC_MAX_INPUT);
					goto call_error;
				}
				else
					*dest[ii] = thr;
			}
			aa+=(1+nb_param);
		}
		else if (strcmp(argv[aa], "-tempo") == 0) {
			int tempo;
			nb_param = 3;
			if (aa+nb_param >= argc) goto call_error;
			dest[0] = &ctx->sl_tempo;
			dest[1] = &ctx->ll_tempo;
			dest[2] = &ctx->dn_tempo;
			for (ii=0; ii<nb_param; ii++) {
				tempo = atoi(argv[aa+1+ii]);
				if (tempo < 0) {
					printf("\n\nERROR: tempo < 0d\n\n");
					goto call_error;
				}
				else
					*dest[ii] = tempo;
			}
			aa+=(1+nb_param);
		}
		else if (strcmp(argv[aa], "-pan_gpio") == 0) {
			if (aa+1 >= argc) goto call_error;
			int gpio;
			gpio = atoi(argv[aa+1]);
			if ((gpio < 0) || (gpio > 120)) {
				printf("\n\nERROR: pan servo GPIO must be \
between 0 and 120\n\n");
				goto call_error;
			}
			else
				ctx->pan.gpio = gpio;
			aa+=2;
		}
		else if (strcmp(argv[aa], "-pan_servo_gain") == 0) {
			if (aa+1 >= argc) goto call_error;
			ctx->pan.gain = atoi(argv[aa+1]);
			aa+=2;
		}
		else if (strcmp(argv[aa], "-pan_range") == 0) {
			int pos;
			nb_param = 2;
			if (aa+nb_param >= argc) goto call_error;
			dest[0] = &ctx->pan.min;
			dest[1] = &ctx->pan.max;
			for (ii=0; ii<nb_param; ii++)
				*dest[ii] = atoi(argv[aa+1+ii]);
			if (ctx->pan.min > ctx->pan.max) {
				printf("\n\nERROR: pan min value (%d deg) \
higher than pan max value (%d deg)\n\n", ctx->pan.min, ctx->pan.max);
				goto call_error;
			}
		}
		else if (strcmp(argv[aa], "-pan_default") == 0) {
			if (aa+1 >= argc) goto call_error;
			int pan;
			pan = atoi(argv[aa+1]);
			if ((pan < ctx->pan.min) || (pan > ctx->pan.max)) {
				printf("\n\nERROR: pan default must be between\
pan min (%ddeg) and pan max (%ddeg)\n\n", ctx->pan.min, ctx->pan.max);
				goto call_error;
			}
			else
				ctx->pan.default_pos = pan;
			aa+=2;
		}
		else if (strcmp(argv[aa], "-pan_darkness") == 0) {
			if (aa+1 >= argc) goto call_error;
			int pan;
			pan = atoi(argv[aa+1]);
			if ((pan < ctx->pan.min) || (pan > ctx->pan.max)) {
				printf("\n\nERROR: pan darkness must be between\
pan min (%ddeg) and pan max (%ddeg)\n\n", ctx->pan.min, ctx->pan.max);
				goto call_error;
			}
			else
				ctx->pan.default_pos = pan;
			aa+=2;
		}
		else if (strcmp(argv[aa], "-tilt_gpio") == 0) {
			if (aa+1 >= argc) goto call_error;
			int gpio;
			gpio = atoi(argv[aa+1]);
			if ((gpio < 0) || (gpio > 120)) {
				printf("\n\nERROR: tilt servo GPIO must be \
between 0 and 120\n\n");
				goto call_error;
			}
			else
				ctx->tilt.gpio = gpio;
			aa+=2;
		}
		else if (strcmp(argv[aa], "-tilt_servo_gain") == 0) {
			if (aa+1 >= argc) goto call_error;
			ctx->tilt.gain = atoi(argv[aa+1]);
			aa+=2;
		}
		else if (strcmp(argv[aa], "-tilt_range") == 0) {
			int pos;
			nb_param = 2;
			if (aa+nb_param >= argc) goto call_error;
			dest[0] = &ctx->tilt.min;
			dest[1] = &ctx->tilt.max;
			for (ii=0; ii<nb_param; ii++)
				*dest[ii] = atoi(argv[aa+1+ii]);
			if (ctx->tilt.min > ctx->tilt.max) {
				printf("\n\nERROR: tilt min value (%d deg) \
higher than tilt max value (%d deg)\n\n", ctx->tilt.min, ctx->tilt.max);
				goto call_error;
			}
		}
		else if (strcmp(argv[aa], "-tilt_default") == 0) {
			if (aa+1 >= argc) goto call_error;
			int tilt;
			tilt = atoi(argv[aa+1]);
			if ((tilt < ctx->tilt.min) || (tilt > ctx->tilt.max)) {
				printf("\n\nERROR: tilt default must be between\
tilt min (%ddeg) and tilt max (%ddeg)\n\n", ctx->tilt.min, ctx->tilt.max);
				goto call_error;
			}
			else
				ctx->tilt.default_pos = tilt;
			aa+=2;
		}
		else if (strcmp(argv[aa], "-tilt_darkness") == 0) {
			if (aa+1 >= argc) goto call_error;
			int tilt;
			tilt = atoi(argv[aa+1]);
			if ((tilt < ctx->tilt.min) || (tilt > ctx->tilt.max)) {
				printf("\n\nERROR: tilt darkness must be between\
tilt min (%ddeg) and tilt max (%ddeg)\n\n", ctx->tilt.min, ctx->tilt.max);
				goto call_error;
			}
			else
				ctx->tilt.default_pos = tilt;
			aa+=2;
		}
		else
			goto call_error;
	}
	return 0;

 call_error:
	printf("\nError. Call syntax is: solar_power [OPTION] ...\n");
	printf("-hemisphere value            : hemisphere in which the panel is\n");
	printf("                               value = north or south\n");
	printf("-adc_channel channel         : the ADC input to measure the solar\n");
	printf("                               panel voltage, 0 <= channel <= 3\n");
	printf("-thresholds sun_thr dark_thr : thresholds in mV to select the \n");
	printf("                               operating mode.\n");
	printf("                               0 <= threshold <= %d\n", ADC_MAX_INPUT);
	printf("-tempo sun_t low_t dark_t    : tempos for each state in seconds:\n");
	printf("                               sun_t for the sunlight state,\n");
	printf("                               low_t for the lowlight state,\n");
	printf("                               dark_t for the darkness state.\n");
	printf("-pan_gpio gpio_number        : GPIO to drive pan servo\n");
	printf("-pan_servo_gain gain_value   : gain to convert pan angle in degrees\n");
	printf("                               into a pulse in microseconds\n");
	printf("-pan_range min max           : min and max pan values in degrees\n");
	printf("-pan_default value           : default pan value in degrees\n");
	printf("-pan_darkness value          : pan angle in degrees for darkness\n");
	printf("-tilt_gpio gpio_number       : GPIO to drive tilt servo\n");
	printf("-tilt_servo_gain gain_value  : gain to convert tilt angle in degrees\n");
	printf("                               into a pulse in microseconds\n");
	printf("-tilt_range min max          : min and max tilt values in degrees\n");
	printf("-tilt_default value          : default tilt in degrees\n");
	printf("-tilt_darkness value         : tilt angle in degrees for darkness\n");

	return -1;
}

int read_panel_voltage(msp_context_t *ctx)
{
	int panel_voltage;
	struct misc_adc adc_data;

	adc_data.channel_count = 1;
	adc_data.channel[0].number = ctx->adc_channel;
	if (ioctl(ctx->adc_fd, MISCIOCADC, (void*) &adc_data) < 0) {
		perror("read_panel_voltage ioctl ADC: ");
		panel_voltage = -1;
	}
	else {
		/* ADC resolution is about 7.5mV/bit */
		panel_voltage = (adc_data.channel[0].value * 15) >> 1;
	}
#ifdef DEBUG
	printf("Vsp = %dmV - pan = %ddeg - tilt = %ddeg\n",
	       panel_voltage, ctx->pan.angle, ctx->tilt.angle);
#endif
	return panel_voltage;
}

int servo_delay(int angle, msp_servo_t *servo) {
	if ((angle < servo->min) || (angle > servo->max)) {
		printf("servo_delay: %d out of range [%d;%d]\n",
		       angle, servo->min, servo->max);
		return 0;
	}
	else {
		return 1500 + angle * servo->gain;
	}
}

void set_servo_angle(msp_context_t *ctx, msp_servo_t *servo, int angle)
{
	struct pxa_servo_chg srv_chg;
	struct pxa_servo_val srv_val;

	if (angle < servo->min) angle = servo->min;
	if (angle > servo->max) angle = servo->max;
	srv_chg.nb = 1;
	srv_chg.val = &srv_val;
	srv_val.id = servo->id;
	srv_val.delay = 1500 + angle * servo->gain;
	if (ioctl(ctx->servo_fd, PXAIOCSRVCHG,
		  (void*) &srv_chg)) {
		perror("set_new_angle ioctl PXAIOCSRVCHG: ");
	}
	else {
		/* wait for rotation end */
		int delay;
		delay = (angle - servo->angle) * 1000000;
		if (delay < 0) delay = -delay;
		delay = delay / SERVO_SPEED;
		if (delay < 250000) delay = 250000;
#ifdef DEBUG2
		printf("set_servo_angle: rotation delay = %d\n", delay);
#endif
		usleep(delay);
		servo->angle = angle;
	}
}

void msp_single_axis_search_max(msp_context_t *ctx, msp_servo_t *servo,
				int angle_min, int angle_max)
{
	int angle, opt_angle, panel_voltage_max;

	if (angle_min < servo->min) angle_min = servo->min;
	if (angle_max > servo->max) angle_max = servo->max;
	angle = angle_min;
	set_servo_angle(ctx, servo, angle);
	panel_voltage_max = 0;
	while (angle <= angle_max) {
		/* wait for panel stabilization before measuring */
		sleep(1);
		ctx->panel_voltage = read_panel_voltage(ctx);
		if (ctx->panel_voltage > panel_voltage_max) {
			panel_voltage_max = ctx->panel_voltage;
			opt_angle = angle;
		}
		angle += SEARCH_STEP;
		set_servo_angle(ctx, servo, angle);
	}
	set_servo_angle(ctx, servo, opt_angle);
}

/* Find the globally optimal panel position by scanning pan and tilt ranges.
   Final position is determined by tracking */
void msp_search_max(msp_context_t *ctx,
		    int pan_min, int pan_max, int tilt_min, int tilt_max)
{
	msp_single_axis_search_max(ctx, &ctx->pan, pan_min, pan_max);
	msp_single_axis_search_max(ctx, &ctx->tilt, tilt_min, tilt_max);
#ifdef DEBUG
	printf("msp_search_max end:\n");
	ctx->panel_voltage = read_panel_voltage(ctx);
#endif
}

void msp_single_axis_search_opt(msp_context_t *ctx, msp_servo_t *servo)
{
	int angle, opt_angle, step, old_panel_voltage;
	int bconstant;
	struct pxa_servo_chg srv_chg;
	struct pxa_servo_val srv_val;

	opt_angle = servo->angle;
	old_panel_voltage = read_panel_voltage(ctx);
	step = -ctx->hemisphere;
	bconstant = 1;
	while (1) {
		angle = servo->angle+step;
		if (angle < servo->min) {
			if (bconstant) {
				/* search the other direction */
				bconstant = 0;
				step = -step;
				angle = servo->min;
			}
			else {
				break;
			}
		}
		if (angle > servo->max) {
			if (bconstant) {
				/* search the other direction */
				bconstant = 0;
				step = -step;
				angle = servo->max;
			}
			else {
				break;
			}
		}
		set_servo_angle(ctx, servo, angle);
		ctx->panel_voltage = read_panel_voltage(ctx);
		if (ctx->panel_voltage > old_panel_voltage) {
			bconstant = 0;
			old_panel_voltage = ctx->panel_voltage;
			opt_angle = servo->angle;
		}
		else if (ctx->panel_voltage < old_panel_voltage) {
			if (bconstant) {
				/* search in the wrong direction */
				bconstant = 0;
				step = -step;
			}
			else {
				break;
			}
		}
	}
	set_servo_angle(ctx, servo, opt_angle);
#ifdef DEBUG
	printf("msp_single_axis_search_opt: opt_angle = %d\n",
	       opt_angle);
#endif
}

/* Find the panel optimal position from a neighbour position */
void msp_search_opt(msp_context_t *ctx)
{
	msp_single_axis_search_opt(ctx, &ctx->pan);
	msp_single_axis_search_opt(ctx, &ctx->tilt);
#ifdef DEBUG
	printf("msp_search_opt end:\n");
	ctx->panel_voltage = read_panel_voltage(ctx);
#endif
}

int msp_init(msp_context_t *ctx)
{
	printf("Starting devices\n");

	struct pxa_servo_add srv_add;

	ctx->adc_fd = open(ADC_DEV_NAME, O_RDWR);
	if(ctx->adc_fd < 0) {
		printf("%s ", ADC_DEV_NAME);
		perror("open failed");
		return -1;
	}

	ctx->servo_fd = open(SERVO_DEV_NAME, O_RDWR);
	if(ctx->servo_fd < 0) {
		printf("%s ", SERVO_DEV_NAME);
		perror("open failed");
		return -1;
	}

	srv_add.gpio = ctx->pan.gpio;
	srv_add.delay = servo_delay(ctx->pan.default_pos, &ctx->pan);
	if (ioctl(ctx->servo_fd, PXAIOCSRVADD, (void*) &srv_add)) {
		perror("Pan servo ioctl PXAIOCSRVADD: ");
		return -1;
	}
	ctx->pan.id = srv_add.id;

	srv_add.gpio = ctx->tilt.gpio;
	srv_add.delay = servo_delay(ctx->tilt.default_pos, &ctx->tilt);
	if (ioctl(ctx->servo_fd, PXAIOCSRVADD, (void*) &srv_add)) {
		perror("Tilt servo ioctl PXAIOCSRVADD: ");
		return -1;
	}
	ctx->tilt.id = srv_add.id;

	/* wait a couple of seconds to let the servos reach their
	   initial position */
	sleep(2);

	/* find a position close to the optimum for tracking */
	ctx->pan.angle = ctx->pan.default_pos;
	ctx->tilt.angle = ctx->tilt.default_pos;
	msp_search_max(ctx,
		       ctx->pan.min, ctx->pan.max,
		       ctx->tilt.min, ctx->tilt.max);
	ctx->state = MSP_STATE_TRACKING;

	return 0;
}

void msp_mainloop(msp_context_t *ctx)
{
	while (1) {
#ifdef DEBUG
	display_context(ctx, 1, "msp_mainloop");
#else
	display_context(ctx, 0, "msp_mainloop");
#endif
		switch (ctx->state) {
		case MSP_STATE_SUNLIGHT :
			sleep(ctx->sl_tempo);
			ctx->state = MSP_STATE_TRACKING;
			break;
		case MSP_STATE_LOWLIGHT :
			sleep(ctx->ll_tempo);
			ctx->state = MSP_STATE_TRACKING;
			break;
		case MSP_STATE_DARKNESS :
			sleep(ctx->dn_tempo);
			ctx->panel_voltage = read_panel_voltage(ctx);
			if (ctx->panel_voltage > ctx->dn_thr)
				ctx->state = MSP_STATE_TRACKING;
			break;
		case MSP_STATE_TRACKING :
			msp_search_opt(ctx);
			ctx->panel_voltage = read_panel_voltage(ctx);
			if (ctx->panel_voltage > ctx->sl_thr)
				ctx->state = MSP_STATE_SUNLIGHT;
			else if (ctx->panel_voltage < ctx->dn_thr) {
				ctx->state = MSP_STATE_DARKNESS;
				/* turn the panel towards sunrise
				   when entering this state */
				set_servo_angle(ctx, &ctx->pan,
						ctx->pan.darkness_pos);
				set_servo_angle(ctx, &ctx->tilt,
						ctx->tilt.darkness_pos);
			}
			else
				ctx->state = MSP_STATE_LOWLIGHT;

			break;
		default :
			printf("msp_mainloop: unknown state %d\n", ctx->state);
		}
	}
}

void msp_stop(msp_context_t *ctx)
{
	printf("Stopping devices\n");

	if (ctx->adc_fd > 0) close (ctx->adc_fd);
	if (ctx->servo_fd > 0) {
		struct pxa_servo_rem srv_rem;
		srv_rem.id = ctx->pan.id;
		ioctl(ctx->servo_fd, PXAIOCSRVREM, (void*) &srv_rem);
		srv_rem.id = ctx->tilt.id;
		ioctl(ctx->servo_fd, PXAIOCSRVREM, (void*) &srv_rem);
		close(ctx->servo_fd);
	}
}

int main(int argc, char *argv[])
{
	msp_context_t ctx;

	printf("\n\n   ***   solar_power   ***\n\n");
	printf("Sample program monitoring a solar panel.\n\n");

	init_defaults(&ctx);
	display_context(&ctx, 1, "Default settings");
	
	if (read_args(&ctx, argc, argv)) exit(1);
	display_context(&ctx, 1, "Settings after args read");
	
	if (msp_init(&ctx)) goto msp_stop_;

	msp_mainloop(&ctx);

 msp_stop_:
	msp_stop(&ctx);

	exit(0);
}
