/*****************************************/
/*                                       */
/*      MBS270 V2 demo server            */
/*                                       */
/* Copyright Mobisense Systems 2008      */
/*                                       */
/* Main source file                      */
/*                                       */
/*****************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/fb.h>
#include <linux/types.h>
#include <linux/videodev2.h>
#include <linux/ov7720.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include "mbs270_demo.h"
#ifdef MDS_WITH_IPP
#include "gradient_computations.h"
#endif

/* Application messages:
   . 0 = none
   . 1 = minimal
   . 2 = details */
#define MDS_VERBOSE 1

#define VIDEO_DEV_NAME "/dev/video0"

#define OV7720_I2CADDR 0x21	/* OV7720 unique I2C address */
static __u32 i2c_addr[] = {
	OV7720_I2CADDR,
	0x48,	/* MT9V032 default I2C address */
	0x4c,	/* MT9V032 alternate I2C address */
	0x58,	/* MT9V032 alternate I2C address */
	0x5c	/* MT9V032 alternate I2C address */
};

#define OV7720_CHIPID	251
#define MT9V032M_CHIPID	45015
#define MT9V032C_CHIPID	45010

/* Possible states
   . Wait for client connexion
   . Stand-by: wait for client request
   . Greyscale image acquisition
   . YCRCB422 image acquisition
   . RGB24 image acquisition
   . Greyscale image acquisition and Gradient computation */
typedef enum {
	MDS_STATE_CNX_WAIT,
	MDS_STATE_STD_BY,
	MDS_STATE_ACQ_IMG_Y8,
	MDS_STATE_ACQ_IMG_YUV422P,
	MDS_STATE_ACQ_IMG_YUYV,
	MDS_STATE_ACQ_IMG_GRB422P,
	MDS_STATE_ACQ_IMG_GRGB,
	MDS_STATE_ACQ_IMG_BAYER,
	MDS_STATE_ACQ_IMG_HGRAD,
	MDS_STATE_ACQ_IMG_VGRAD,
	MDS_STATE_ACQ_IMG_EDGE
} mds_state_t;

#ifndef timersub
/* Some includes doesn't contain this macro */
#define	timersub(a, b, result)						\
	do {								\
		(result)->tv_sec = (a)->tv_sec - (b)->tv_sec;		\
		(result)->tv_usec = (a)->tv_usec - (b)->tv_usec;	\
		if ((result)->tv_usec < 0) {				\
			--(result)->tv_sec;				\
			(result)->tv_usec += 1000000;			\
		}							\
	} while (0)
#endif

/* A small function to align a pointer to the next page boundary */
/* runs only on 32-bits machines                                 */
unsigned int page_align(unsigned int ptr)
{
	unsigned int page_size;
	page_size = getpagesize();
	return ((ptr + page_size - 1) & ~(page_size - 1));
}


struct image_buffer {
	void *start;
	size_t length;
};

struct demo_context {
	__u32 width;
	__u32 height;
	__u32 frame_size;
	__u32 pixelformat;
	__u32 i2c_addr;
	__u32 video_input;
	int   res;
	int sensor;
	int   vfd;
	int   lsock, csock;
	mds_state_t state;
	int   nr_buffers;
	struct image_buffer *buffers;
#ifdef MDS_WITH_IPP
	mds_grad_attr_t grad_h, grad_v;
	mds_edge_attr_t edge;
#endif
};


/* mds_init_video initializes video capture and allocates buffers */
int mds_init_video(struct demo_context *dctx)
{
	int ret, ii;
	struct v4l2_format format;
	struct v4l2_requestbuffers reqbuf;
	struct v4l2_buffer buffer;

	ret = 0;

	dctx->frame_size = dctx->width * dctx->height;
	printf("mds_init_video: frame_size=%u - width=%u - height=%u\n",
	       dctx->frame_size, dctx->width, dctx->height);
	format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	format.fmt.pix.width = dctx->width;
	format.fmt.pix.height = dctx->height;
	format.fmt.pix.pixelformat = dctx->pixelformat;
	format.fmt.pix.field = V4L2_FIELD_NONE;

	if (ioctl(dctx->vfd, VIDIOC_S_FMT, &format) != 0) {
		perror("ioctl VIDIOC_S_FMT");
		return -1;
	}

	/* request buffers to store grabbed image */
	reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	reqbuf.memory = V4L2_MEMORY_MMAP;
	reqbuf.count = dctx->nr_buffers;

	if (ioctl(dctx->vfd, VIDIOC_REQBUFS, &reqbuf) < 0) {
		if (errno == EINVAL)
			printf ("Video capturing or mmap-streaming is \
not supported\n");
		else
			perror ("ioctl VIDIOC_REQBUFS");
		return -1;
	}

	if (reqbuf.count < dctx->nr_buffers) {
		printf("   !!! Only %d buffers available\n", reqbuf.count);
		dctx->nr_buffers = reqbuf.count;
	}

	dctx->buffers = calloc(dctx->nr_buffers, sizeof (*dctx->buffers));
	if (dctx->buffers == NULL) {
		perror("calloc dctx->buffers");
		return -1;
	}

	for (ii=0; ii<dctx->nr_buffers; ii++) {
		/* query buffer to allocate and mmap it */
		memset(&buffer, 0, sizeof (buffer));
		buffer.type = reqbuf.type;
		buffer.memory = V4L2_MEMORY_MMAP;
		buffer.index = ii;

		if (ioctl(dctx->vfd, VIDIOC_QUERYBUF, &buffer) < 0) {
			perror ("ioctl VIDIOC_QUERYBUF");
			return -1;
		}

#if MDS_VERBOSE > 0
		printf("buffer %u: length = %u, bytesused = %u\n",
		       buffer.index, buffer.length, buffer.bytesused);
#endif

		dctx->buffers[ii].length = buffer.length; /* save for munmap */
		dctx->buffers[ii].start = mmap (NULL, buffer.length,
						PROT_READ | PROT_WRITE,
						MAP_SHARED,
						dctx->vfd, buffer.m.offset);
		if (MAP_FAILED == dctx->buffers[ii].start) {
			perror ("mmap");
			return -1;
		}
	}

	memset(&buffer, 0, sizeof (buffer));
	buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buffer.memory = V4L2_MEMORY_MMAP;

	for (ii=0; ii<dctx->nr_buffers; ii++) {
		buffer.index = ii;
		ret = ioctl(dctx->vfd, VIDIOC_QBUF, &buffer);
		if (ret < 0) {
			perror ("ioctl VIDIOC_QBUF");
			return ret;
		}

#if MDS_VERBOSE > 0
		printf("buffer %u: bytesused = %u\n",
		       buffer.index, buffer.bytesused);
#endif
	}

	ret = ioctl(dctx->vfd, VIDIOC_STREAMON, &buffer.type);
	if (ret < 0) {
		perror ("ioctl VIDIOC_STREAMON");
		return ret;
	}

	for (ii=0; ii<dctx->nr_buffers; ii++) {
		/* dequeue initial images */
		buffer.index = ii;
		ret = ioctl(dctx->vfd, VIDIOC_DQBUF, &buffer);
		if (ret < 0) {
			if (errno == EAGAIN)
				printf("Image not grabbed yet.\n");
			else
				perror("ioctl VIDIOC_DQBUF");
			return ret;
		}
	}

	printf("Video init OK\n\n");

	return 0;
}


/* mds_stop_video stops video capture and releases buffers */
int mds_stop_video(struct demo_context *dctx)
{
	int ret, ii;
	enum v4l2_buf_type type;

	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	if (ioctl(dctx->vfd, VIDIOC_STREAMOFF, &type) != 0) {
		perror ("ioctl VIDIOC_STREAMOFF");
		return -1;
	}

	if (dctx->buffers != NULL) {
		for (ii=0; ii<dctx->nr_buffers; ii++) {
			if ((dctx->buffers[ii].start != MAP_FAILED) &&
			    (dctx->buffers[ii].length > 0))
				munmap(dctx->buffers[ii].start,
				       dctx->buffers[ii].length);
		}
		free(dctx->buffers);
	}

	return 0;
}


int mds_acquire_image(struct demo_context *dctx)
{
	int ret;
	struct v4l2_buffer buffer;
	struct timeval tvs, tve, tvd;
	
	ret = 0;
	
	buffer.index = 0;
	buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buffer.flags = 0;
	buffer.memory = V4L2_MEMORY_MMAP;

#if MDS_VERBOSE > 1
	gettimeofday(&tvs, NULL);
#endif

	ret = ioctl(dctx->vfd, VIDIOC_QBUF, &buffer);
	if (ret < 0) {
		perror("ioctl VIDIOC_QBUF");
		goto warn_acq_img_failed;
	}

	ret = ioctl(dctx->vfd, VIDIOC_DQBUF, &buffer);
	if (ret < 0) {
		if (errno == EAGAIN)
			printf("Image not grabbed yet.\n");
		else {
			perror("ioctl VIDIOC_DQBUF");
			goto warn_acq_img_failed;
		}
	}

#if MDS_VERBOSE > 1
	if (ret == 0) {
		gettimeofday(&tve, NULL);
		timersub(&tve, &tvs, &tvd);
		printf("frame %u acquisition: start=%ld.%06d  stop=%ld.%06d, \
dt=%ld.%06d\n",
		       buffer.index,
		       tvs.tv_sec, tvs.tv_usec,
		       tve.tv_sec, tve.tv_usec,
		       tvd.tv_sec, tvd.tv_usec);
	}
#endif

	return 0;

 warn_acq_img_failed:
	ret = ret;
	char strmsg[32];
	md_msghdr_t	msghdr;
	msghdr.type = MD_MSG_STRING;
	strcpy(strmsg,"Image Acquisition failed.");
	msghdr.bodylen = strlen(strmsg);
	mds_snd_msg(dctx->csock, (void*)&msghdr, sizeof(msghdr));
	mds_snd_msg(dctx->csock, (void*)strmsg, msghdr.bodylen);
	return ret;
}

#ifdef MDS_WITH_IPP
int mds_init_grad(struct demo_context *dctx)
{
	int ret;

	/* setup gradient buffers */
	ret = 0;
	switch (dctx->state) {
	case MDS_STATE_ACQ_IMG_HGRAD :
		ret = mds_init_hgrad(dctx->width, dctx->height,
				     &dctx->grad_h);
		break;
	case MDS_STATE_ACQ_IMG_VGRAD :
		ret = mds_init_vgrad(dctx->width, dctx->height,
				     &dctx->grad_v);
		break;
	case MDS_STATE_ACQ_IMG_EDGE :
		ret = mds_init_edge(dctx->width, dctx->height, &dctx->edge);
		break;
	}

	return ret;
}

void mds_stop_grad(struct demo_context *dctx)
{
	/* release gradient buffers */
	switch (dctx->state) {
	case MDS_STATE_ACQ_IMG_HGRAD :
		mds_deinit_hgrad(&dctx->grad_h);
		break;
	case MDS_STATE_ACQ_IMG_VGRAD :
		mds_deinit_vgrad(&dctx->grad_v);
		break;
	case MDS_STATE_ACQ_IMG_EDGE :
		mds_deinit_edge(&dctx->edge);
		break;
	}
}
#endif


int mds_rec_msg(int sock, void* buf, int len)
{
	int nr_bytes;
	int retval = 0;
	int nr_bytes_to_receive = len;
	void* pbuf = buf;

	while (nr_bytes_to_receive > 0) {
		nr_bytes = recv(sock, pbuf, nr_bytes_to_receive, 0);
		if (nr_bytes == -1) {
			perror("recv");
			retval = -1;
			break;
		} else {
			nr_bytes_to_receive -= nr_bytes;
			pbuf += nr_bytes;
		}
	}
	return retval;
}


int mds_snd_msg(int sock, void* buf, int len)
{
	int nr_bytes;
	int retval = 0;
	int nr_bytes_to_send = len;
	void* pbuf = buf;

	while (nr_bytes_to_send > 0) {
		nr_bytes = send(sock, pbuf, nr_bytes_to_send, 0);
		if (nr_bytes == -1) {
			perror("send");
			retval = -1;
			break;
		} else {
			nr_bytes_to_send -= nr_bytes;
			pbuf += nr_bytes;
		}
	}
	return retval;
}


void mds_close_socket(struct demo_context *dctx)
{
	close(dctx->csock);
	dctx->csock = -1;
	dctx->state = MDS_STATE_CNX_WAIT;
}

int mds_init(struct demo_context *dctx)
{
	int ii, nb_i2c_addr;
	struct v4l2_chip_ident chip_id;
	struct v4l2_control ctrl;
	struct v4l2_format format;
	struct v4l2_buffer buffer;

	/* only one buffer in this application */
	dctx->nr_buffers = 1;

	dctx->csock = -1;
	dctx->vfd = -1;
	dctx->buffers = NULL;

	/* Open video device */
	dctx->vfd = open(VIDEO_DEV_NAME, O_RDWR);
	if (dctx->vfd == -1) {
		perror("open");
		return -1;
	}
#if MDS_VERBOSE > 0
	else printf("%s open succeeded.\n", VIDEO_DEV_NAME);
#endif

	/* select input 0, the only one */
	dctx->video_input = 0;

	if (ioctl(dctx->vfd, VIDIOC_S_INPUT, &dctx->video_input) != 0) {
		perror("ioctl VIDIOC_S_INPUT");
		return -1;
	}

	/* detect sensor chip ID */
	nb_i2c_addr = sizeof(i2c_addr) / sizeof(i2c_addr[0]);
	dctx->i2c_addr = -1;
	chip_id.match_type = V4L2_CHIP_MATCH_I2C_ADDR;
	for (ii=0; ii<nb_i2c_addr; ii++) {
		chip_id.match_chip = i2c_addr[ii];
		if (ioctl(dctx->vfd, VIDIOC_G_CHIP_IDENT, &chip_id) == 0) {
			dctx->i2c_addr = chip_id.match_chip;
			break;
		}
	}
	if (dctx->i2c_addr == -1) {
		printf("No sensor found. Make sure camera is connected.\n");
		return -1;
	}
	switch (chip_id.ident) {
	case OV7720_CHIPID :
		dctx->sensor = MD_SENSOR_OV7720;
		break;
	case MT9V032M_CHIPID :
		dctx->sensor = MD_SENSOR_MT9V032M;
		break;
	case MT9V032C_CHIPID :
		dctx->sensor = MD_SENSOR_MT9V032C;
		break;
	default :
		printf("Unknown chip ID %d\n", chip_id.ident);
		return -1;
	}

	/* Socket init */
	struct sockaddr_in lsin;

	dctx->lsock = socket(PF_INET, SOCK_STREAM, 0);
	if (dctx->lsock == -1) {
		perror("socket");
		return -1;
	}

	lsin.sin_addr.s_addr	= INADDR_ANY;
	lsin.sin_family	= AF_INET;
	lsin.sin_port		= htons(4444);
	if (bind(dctx->lsock, (struct sockaddr *)&lsin, sizeof(lsin)) != 0) {
		perror("bind");
		return -1;
	}

	if (listen(dctx->lsock, 0) != 0) {
		perror("listen");
		return -1;
	}
	else {
		dctx->state = MDS_STATE_CNX_WAIT;
#if MDS_VERBOSE > 0
		printf("Listening socket created.\n");
#endif
	}

	return 0;
}


void mds_send_answer(struct demo_context *dctx,
		     md_msghdr_t* msghdr,
		     void* body)
{
	int ret = 0;
	if (mds_snd_msg(dctx->csock, (void*) msghdr, sizeof(*msghdr)) == -1)
		ret = -1;
	else if (mds_snd_msg(dctx->csock, body, msghdr->bodylen) == -1)
		ret = -1;
	if (ret == -1) mds_close_socket(dctx);
}


/* Send answer to planar format request:
   because of possible discontinuity between buffers of each component,
   we need to send them separately */
void mds_send_answer_pl(struct demo_context *dctx,
			md_msghdr_t* msghdr,
			void* pl_buf)
{
	int ret;
	void *ptr;

	ret = mds_snd_msg(dctx->csock, (void*) msghdr, sizeof(*msghdr));
	if (ret == 0) {
		ptr = pl_buf;
		ret = mds_snd_msg(dctx->csock, ptr, dctx->frame_size);
	}
	if (ret == 0) {
		ptr += page_align(dctx->frame_size);
		ret = mds_snd_msg(dctx->csock, ptr, dctx->frame_size/2);
	}
	if (ret == 0) {
		ptr += page_align(dctx->frame_size/2);
		ret = mds_snd_msg(dctx->csock, ptr, dctx->frame_size/2);
	}
	if (ret != 0) mds_close_socket(dctx);
}


void mds_process_get_reg_val_request(struct demo_context *dctx)
{
	int reg_addr;
	struct v4l2_register reg;

	if (mds_rec_msg(dctx->csock, (void*) &reg_addr, sizeof(reg_addr)) == 0) {
		reg.match_type = V4L2_CHIP_MATCH_I2C_ADDR;
		reg.match_chip = dctx->i2c_addr;
		reg.reg = reg_addr;
		if (ioctl(dctx->vfd, VIDIOC_DBG_G_REGISTER, &reg))
			perror("ioctl VIDIOC_DBG_G_REGISTER");
		else {
			md_msghdr_t msghdr;
			int reg_val = reg.val;
			msghdr.type = MD_MSG_REG_VAL;
			msghdr.bodylen = sizeof(reg_val);
			mds_send_answer(dctx, &msghdr, (void*)&reg_val);
		}
	}
}


void mds_process_set_reg_val_request(struct demo_context *dctx)
{
	struct {int addr, val;} request;
	struct v4l2_register reg;

	if (mds_rec_msg(dctx->csock, (void*) &request, sizeof(request)) == 0) {
		reg.match_type = V4L2_CHIP_MATCH_I2C_ADDR;
		reg.match_chip = dctx->i2c_addr;
		reg.reg = request.addr;
		reg.val = request.val;
		if (ioctl(dctx->vfd, VIDIOC_DBG_S_REGISTER, &reg))
			perror("ioctl VIDIOC_DBG_S_REGISTER");
	}
}


int mds_set_sensor_resolution(struct demo_context *dctx, int res)
{
	int ret;
	struct v4l2_control ctrl;

	ret = 0;

	switch (res) {
	case MD_RES_QVGA :
		dctx->width = 320;
		dctx->height = 240;
		break;
	case MD_RES_VGA :
		dctx->width = 640;
		dctx->height = 480;
		break;
	case MD_RES_QWVGA :
		dctx->width = 376;
		dctx->height = 240;
		break;
	case MD_RES_WVGA :
		dctx->width = 752;
		dctx->height = 480;
		break;
	default :
		printf("Unknown resolution %d\n", res);
		return -1;
	}

/* 	ctrl.id = OV7720_CID_QVGA; */
/* 	ctrl.value = dctx->qvga; */
/* 	ret = ioctl(dctx->vfd, VIDIOC_S_CTRL, &ctrl); */
/* 	if (ret != 0) { */
/* 		perror("ioctl VIDIOC_S_CTRL OV7720_CID_QVGA"); */
/* 		return ret; */
/* 	} */

	if (dctx->state != MDS_STATE_STD_BY) {
		/* restart acquisition with well sized buffers */
		if (mds_stop_video(dctx) != 0) return -1;
#ifdef MDS_WITH_IPP
		mds_stop_grad(dctx);
#endif
		if (mds_init_video(dctx) != 0) return -1;
#ifdef MDS_WITH_IPP
		if (mds_init_grad(dctx) != 0) return -1;
#endif
	}

	return 0;
}


void mds_process_request(struct demo_context *dctx)
{
	mcd_rq_t rq;
	if (mds_rec_msg(dctx->csock, (void*) &rq, sizeof(rq)) == 0) {
		if (rq == MD_RQ_STDBY) {
			/* stop image acquisition */
			mds_stop_video(dctx);
#ifdef MDS_WITH_IPP
			mds_stop_grad(dctx);
#endif
			dctx->state = MDS_STATE_STD_BY;
		}
		else if (rq == MD_RQ_GET_REG_VAL) {
			mds_process_get_reg_val_request(dctx);
		}
		else if (rq == MD_RQ_SET_REG_VAL) {
			mds_process_set_reg_val_request(dctx);
		}
		else if (rq == MD_RQ_RES_VGA) {
			mds_set_sensor_resolution(dctx, MD_RES_VGA);
		}
		else if (rq == MD_RQ_RES_QVGA) {
			mds_set_sensor_resolution(dctx, MD_RES_QVGA);
		}
		else
			/* ignore other requests */
			printf("request %d ignored\n", rq);
	}
	else {
		/* connection lost => close socket */
		mds_close_socket(dctx);
	}
}


void mds_check_request(struct demo_context *dctx)
{
	/* check for incoming requests */
	struct timeval tv;
	fd_set rfds, efds;
	
	FD_ZERO(&rfds); FD_SET(dctx->csock, &rfds);
	FD_ZERO(&efds); FD_SET(dctx->csock, &efds);
	tv.tv_sec = 0; tv.tv_usec = 0;	/* non blocking */
#if MDS_VERBOSE > 1
	printf("Entering select ...\n");
#endif
	int retval = select(dctx->csock+1, &rfds, NULL, &efds, &tv);
	if (retval == -1) perror("select");
	else {
#if MDS_VERBOSE > 1
		printf("After select, retval = %d\n", retval);
#endif
		if (retval != 0) {
			if (FD_ISSET(dctx->csock, &efds)) {
				/* socket event happened = connection lost
				   => close socket */
				mds_close_socket(dctx);
			}
			else if (FD_ISSET(dctx->csock, &rfds)) {
				/* request received */
				mds_process_request(dctx);
			}
		}
	}
}


void mds_send_initdata(struct demo_context *dctx)
{
	md_msghdr_t msghdr;
	md_initdata_t initdata;

	msghdr.type = MD_MSG_INITDATA;
	msghdr.bodylen = sizeof(initdata);
	initdata.sensor = dctx->sensor;

	/* sensor specific capabilities (goes beyond V4L2 capabilities) */
	switch (dctx->sensor) {
	case MD_SENSOR_OV7720 :
		initdata.resolutions = MD_RES_QVGA | MD_RES_VGA;
		initdata.default_res = MD_RES_QVGA;
		initdata.formats =
			MD_FMT_YUV422P | MD_FMT_YUYV |
			MD_FMT_GRB422P | MD_FMT_GRGB | MD_FMT_Y8 |
			MD_FMT_HGRAD | MD_FMT_VGRAD | MD_FMT_EDGE;
		initdata.default_fmt = MD_FMT_YUV422P;
		break;
	case MD_SENSOR_MT9V032M :
		initdata.resolutions = MD_RES_WVGA | MD_RES_VGA;
		initdata.default_res = MD_RES_WVGA;
		initdata.formats =
			MD_FMT_Y8 | MD_FMT_HGRAD | MD_FMT_VGRAD | MD_FMT_EDGE;
		initdata.default_fmt = MD_FMT_Y8;
		break;
	case MD_SENSOR_MT9V032C :
		initdata.resolutions = MD_RES_WVGA | MD_RES_VGA;
		initdata.default_res = MD_RES_WVGA;
		initdata.formats     = MD_FMT_BAYER8;
		initdata.default_fmt = MD_FMT_BAYER8;
		break;
	}

	mds_send_answer(dctx, &msghdr, (void*) &initdata);
}


int mds_demo(struct demo_context *dctx)
{
	struct sockaddr_in csin;
	socklen_t sinsize;
	mcd_rq_t rq;
	static md_msghdr_t msghdr;
	struct timeval tvs, tve, tvd;
	struct v4l2_control ctrl;

	switch (dctx->state) {
	case MDS_STATE_CNX_WAIT :
		sinsize = sizeof(csin);
		printf("Waiting for connection ...\n");
		dctx->csock = accept(dctx->lsock, (struct sockaddr *)&csin,
				     &sinsize);
		if (dctx->csock == -1) {
			perror("accept");
			return -1;
		}
	     
		printf("Connected with client.\n");
		dctx->state = MDS_STATE_STD_BY;
		mds_send_initdata(dctx);
		break;
	
	case MDS_STATE_STD_BY :
#if MDS_VERBOSE > 0
		printf("Stand-by, waiting for request ...\n");
#endif
		if (mds_rec_msg(dctx->csock, (void*) &rq, sizeof(rq)) != 0) {
			/* connection lost => close socket */
			close(dctx->csock);
			dctx->csock = -1;
			dctx->state = MDS_STATE_CNX_WAIT;
		}
		else {
			switch (rq) {
			case MD_RQ_QUIT :
				/* stop srv_demo */
				return 1;
				
			case MD_RQ_STDBY :
				/* nothing to do */
				break;
				
			case MD_RQ_IMG_Y8 :
				/* start greyscale image acquisition */
#if MDS_VERBOSE > 0
				printf("   >>>   Greyscale image request\n");
#endif
				if (dctx->sensor == MD_SENSOR_OV7720)
					dctx->pixelformat = V4L2_PIX_FMT_YUV422P;
				else if (dctx->sensor == MD_SENSOR_MT9V032M)
					dctx->pixelformat = V4L2_PIX_FMT_GREY;
				if (mds_init_video(dctx) != 0) return -1;
				printf("init Y8 OK\n");
				dctx->state = MDS_STATE_ACQ_IMG_Y8;
				break;
				
			case MD_RQ_IMG_YUV422P :
				/* start planar YUV422 image acquisition */
#if MDS_VERBOSE > 0
				printf("   >>>   Planar YUV422 image request\n");
#endif
				dctx->pixelformat = V4L2_PIX_FMT_YUV422P;
				if (mds_init_video(dctx) != 0) return -1;
				dctx->state = MDS_STATE_ACQ_IMG_YUV422P;
				break;
				
			case MD_RQ_IMG_YUYV :
				/* start YUYV image acquisition */
#if MDS_VERBOSE > 0
				printf("   >>>   YUYV image request\n");
#endif
				dctx->pixelformat = V4L2_PIX_FMT_YUYV;
				if (mds_init_video(dctx) != 0) return -1;
				dctx->state = MDS_STATE_ACQ_IMG_YUYV;
				break;
				
			case MD_RQ_IMG_GRB422P :
				/* start planar GRB422 image acquisition */
#if MDS_VERBOSE > 0
				printf("   >>>   Planar GRB422 image request\n");
#endif
				dctx->pixelformat = V4L2_PIX_FMT_GRB422P;
				if (mds_init_video(dctx) != 0) return -1;
				dctx->state = MDS_STATE_ACQ_IMG_GRB422P;
				break;
				
			case MD_RQ_IMG_GRGB :
				/* start GRGB image acquisition */
#if MDS_VERBOSE > 0
				printf("   >>>   GRGB image request\n");
#endif
				dctx->pixelformat = V4L2_PIX_FMT_GRGB;
				if (mds_init_video(dctx) != 0) return -1;
				dctx->state = MDS_STATE_ACQ_IMG_GRGB;
				break;
				
			case MD_RQ_IMG_BAYER :
				/* start Bayer image acquisition */
#if MDS_VERBOSE > 0
				printf("   >>>   Bayer image request\n");
#endif
				dctx->pixelformat = V4L2_PIX_FMT_SBGGR8;
				if (mds_init_video(dctx) != 0) return -1;
				dctx->state = MDS_STATE_ACQ_IMG_BAYER;
				break;
				
#ifdef MDS_WITH_IPP
			case MD_RQ_IMG_HGRAD :
				/* start horizontal gradient image computation */
#if MDS_VERBOSE > 0
				printf("   >>>   Horizontal gradient image request\n");
#endif
				if (dctx->sensor == MD_SENSOR_OV7720)
					dctx->pixelformat = V4L2_PIX_FMT_YUV422P;
				else if (dctx->sensor == MD_SENSOR_MT9V032M)
					dctx->pixelformat = V4L2_PIX_FMT_GREY;
				if (mds_init_video(dctx) != 0) return -1;
				if (mds_init_hgrad(dctx->width, dctx->height,
						   &dctx->grad_h) != 0)
					return -1;
				dctx->state = MDS_STATE_ACQ_IMG_HGRAD;
				break;
				
			case MD_RQ_IMG_VGRAD :
				/* start vertical gradient image computation */
#if MDS_VERBOSE > 0
				printf("   >>>   Vertical gradient image request\n");
#endif
				if (dctx->sensor == MD_SENSOR_OV7720)
					dctx->pixelformat = V4L2_PIX_FMT_YUV422P;
				else if (dctx->sensor == MD_SENSOR_MT9V032M)
					dctx->pixelformat = V4L2_PIX_FMT_GREY;
				if (mds_init_video(dctx) != 0) return -1;
				if (mds_init_vgrad(dctx->width, dctx->height,
						   &dctx->grad_v) != 0)
					return -1;
				dctx->state = MDS_STATE_ACQ_IMG_VGRAD;
				break;
				
			case MD_RQ_IMG_EDGE :
				/* start edge image computation */
#if MDS_VERBOSE > 0
				printf("   >>>   Edge image request\n");
#endif
				if (dctx->sensor == MD_SENSOR_OV7720)
					dctx->pixelformat = V4L2_PIX_FMT_YUV422P;
				else if (dctx->sensor == MD_SENSOR_MT9V032M)
					dctx->pixelformat = V4L2_PIX_FMT_GREY;
				if (mds_init_video(dctx) != 0) return -1;
				if (mds_init_edge(dctx->width, dctx->height,
						  &dctx->edge) != 0)
					return -1;
				dctx->state = MDS_STATE_ACQ_IMG_EDGE;
				break;
#endif
				
			case MD_RQ_GET_REG_VAL :
				mds_process_get_reg_val_request(dctx);
				break;

			case MD_RQ_SET_REG_VAL :
				mds_process_set_reg_val_request(dctx);
				break;
				
			case MD_RQ_RES_QVGA :
				mds_set_sensor_resolution(dctx,
							  MD_RES_QVGA);
				break;
				
			case MD_RQ_RES_VGA :
				mds_set_sensor_resolution(dctx,
							  MD_RES_VGA);
				break;
				
			case MD_RQ_RES_QWVGA :
				mds_set_sensor_resolution(dctx,
							  MD_RES_QWVGA);
				break;
				
			case MD_RQ_RES_WVGA :
				mds_set_sensor_resolution(dctx,
							  MD_RES_WVGA);
				break;
				
			default :
				printf("   !!!   unknown request %d   !!!\n",
				       rq);
			}
		}
		break;
	
	case MDS_STATE_ACQ_IMG_Y8 :
		if (mds_acquire_image(dctx) == 0) {
			msghdr.type = MD_MSG_IMG_Y8;
			/* send only the Y component */
			msghdr.bodylen = dctx->frame_size;
			mds_send_answer(dctx, &msghdr, dctx->buffers[0].start);
#if MDS_VERBOSE > 0
			printf("."); fflush(stdout);
#endif
		}
		if (dctx->state == MDS_STATE_ACQ_IMG_Y8)
			mds_check_request(dctx);
		break;
	
	case MDS_STATE_ACQ_IMG_YUV422P :
		if (mds_acquire_image(dctx) == 0) {
			msghdr.type = MD_MSG_IMG_YUV422P;
			/* 2 bytes per pixel with this format */
			msghdr.bodylen = dctx->frame_size*2;
			/* Beware of possible discontinuity between buffers
			   of each component */
			mds_send_answer_pl(dctx, &msghdr,
					   dctx->buffers[0].start);
#if MDS_VERBOSE > 0
			printf("."); fflush(stdout);
#endif
		}
		if (dctx->state == MDS_STATE_ACQ_IMG_YUV422P)
			mds_check_request(dctx);
		break;
	
	case MDS_STATE_ACQ_IMG_YUYV :
		if (mds_acquire_image(dctx) == 0) {
			msghdr.type = MD_MSG_IMG_YUYV;
			msghdr.bodylen = dctx->buffers[0].length;
			mds_send_answer(dctx, &msghdr, dctx->buffers[0].start);
#if MDS_VERBOSE > 0
			printf("."); fflush(stdout);
#endif
		}
		if (dctx->state == MDS_STATE_ACQ_IMG_YUYV)
			mds_check_request(dctx);
		break;
	
	case MDS_STATE_ACQ_IMG_GRB422P :
		if (mds_acquire_image(dctx) == 0) {
			msghdr.type = MD_MSG_IMG_GRB422P;
			/* 2 bytes per pixel with this format */
			msghdr.bodylen = dctx->frame_size*2;
			/* Beware of possible discontinuity between buffers
			   of each component */
			mds_send_answer_pl(dctx, &msghdr,
					   dctx->buffers[0].start);
#if MDS_VERBOSE > 0
			printf("."); fflush(stdout);
#endif
		}
		if (dctx->state == MDS_STATE_ACQ_IMG_GRB422P)
			mds_check_request(dctx);
		break;
	
	case MDS_STATE_ACQ_IMG_GRGB :
		if (mds_acquire_image(dctx) == 0) {
			msghdr.type = MD_MSG_IMG_GRGB;
			msghdr.bodylen = dctx->buffers[0].length;
			mds_send_answer(dctx, &msghdr, dctx->buffers[0].start);
#if MDS_VERBOSE > 0
			printf("."); fflush(stdout);
#endif
		}
		if (dctx->state == MDS_STATE_ACQ_IMG_GRGB)
			mds_check_request(dctx);
		break;
	
	case MDS_STATE_ACQ_IMG_BAYER :
		if (mds_acquire_image(dctx) == 0) {
			msghdr.type = MD_MSG_IMG_BAYER;
			msghdr.bodylen = dctx->frame_size;
			mds_send_answer(dctx, &msghdr, dctx->buffers[0].start);
#if MDS_VERBOSE > 0
			printf("."); fflush(stdout);
#endif
		}
		if (dctx->state == MDS_STATE_ACQ_IMG_BAYER)
			mds_check_request(dctx);
		break;
	
#ifdef MDS_WITH_IPP
	case MDS_STATE_ACQ_IMG_HGRAD :
		if (mds_acquire_image(dctx) == 0) {
			if (mds_compute_hgrad(dctx->buffers[0].start,
					      &dctx->grad_h) == 0) {
				msghdr.type = MD_MSG_IMG_HGRAD;
				msghdr.bodylen = dctx->grad_h.bufSize;
				mds_send_answer(dctx, &msghdr,
						(void*)dctx->grad_h.resBuf);
#if MDS_VERBOSE > 0
				printf("."); fflush(stdout);
#endif
			}
		}
		if (dctx->state == MDS_STATE_ACQ_IMG_HGRAD)
			mds_check_request(dctx);
		break;
	
	case MDS_STATE_ACQ_IMG_VGRAD :
		if (mds_acquire_image(dctx) == 0) {
			if (mds_compute_vgrad(dctx->buffers[0].start,
					      &dctx->grad_v) == 0) {
				msghdr.type = MD_MSG_IMG_VGRAD;
				msghdr.bodylen = dctx->grad_v.bufSize;
				mds_send_answer(dctx, &msghdr,
						(void*)dctx->grad_v.resBuf);
#if MDS_VERBOSE > 0
				printf("."); fflush(stdout);
#endif
			}
		}
		if (dctx->state == MDS_STATE_ACQ_IMG_VGRAD)
			mds_check_request(dctx);
		break;
	
	case MDS_STATE_ACQ_IMG_EDGE :
		if (mds_acquire_image(dctx) == 0) {
			if (mds_compute_edge(dctx->buffers[0].start,
					     &dctx->edge) == 0) {
				msghdr.type = MD_MSG_IMG_EDGE;
				msghdr.bodylen = dctx->edge.bufSize;
				mds_send_answer(dctx, &msghdr,
						(void*)dctx->edge.resBuf);
#if MDS_VERBOSE > 0
				printf("."); fflush(stdout);
#endif
			}
		}
		if (dctx->state == MDS_STATE_ACQ_IMG_EDGE)
			mds_check_request(dctx);
		break;
#endif
	
	default :
		printf("   !!!   unknown state %d   !!!\n", dctx->state);
	}
	
	return 0;
}


void mds_arret(struct demo_context *dctx)
{
	if (dctx->vfd != -1) close(dctx->vfd);
	if (dctx->csock != -1) close(dctx->csock);
	if (dctx->lsock != -1) close(dctx->lsock);
}


int main(void)
{
	struct demo_context dctx;
	int retval = mds_init(&dctx);
	
	while (retval == 0) {
		retval = mds_demo(&dctx);
	}
	
	mds_arret(&dctx);
}
