#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/videodev2.h>

#define VIDEO_DEV_NAME "/dev/video0"
#define IMAGE_DIR "/var/tmp"
#define IMAGE_Y_NAME "image_y.pgm"
#define IMAGE_U_NAME "image_u.pgm"
#define IMAGE_V_NAME "image_v.pgm"
#define IMAGE_G_NAME "image_g.pgm"
#define IMAGE_R_NAME "image_r.pgm"
#define IMAGE_B_NAME "image_b.pgm"
#define IMAGE_NAME   "image.ppm"
#define IMAGE_BAYER_NAME   "image_bayer.ppm"

#define OV7720_AGC		0x00
#define OV7720_AECH		0x08
#define OV7720_COM2		0x09
#define OV7720_COM3		0x0C
#define OV7720_AEC		0x10
#define OV7720_CLKRC		0x11
#define OV7720_COM8		0x13
#define OV7720_COM9		0x14
#define OV7720_HSTART		0x0017
#define OV7720_HSIZE		0x0018
#define OV7720_VSTART		0x0019
#define OV7720_VSIZE		0x001A
#define OV7720_LUMHTH		0x0030
#define OV7720_LUMLTH		0x0031
#define OV7720_HREF		0x0032

#define GREY_STR		"GREY"
#define Y16_STR			"Y16"
#define YUYV_STR		"YUYV"
#define UYVY_STR		"UYVY"
#define YUV422P_STR		"YUV422P"
#define GRGB_STR		"GRGB"
#define GRB422P_STR		"GRB422P"
#define SBGGR8_STR		"SBGGR8"

//#define SAVE_SBGGR8_AS_Y8

/* A small function to align a pointer to the (next) page boundary */
/* runs only on 32-bits machines                                   */
unsigned int page_align(unsigned int ptr)
{
	unsigned int page_size;
	page_size = getpagesize();
	return ((ptr + page_size - 1) & ~(page_size - 1));
}


struct image_buffer {
	void *start;
	size_t length;
};

struct grab_context {
	__u32 width;
	__u32 height;
	__u32 pixelformat;
	__u32 i2c_addr;
	int   vfd;
	int   nr_buffers;
	struct image_buffer *buffers;
};

#define OV7720_I2CADDR	0x21	/* OV7720 unique I2C address */
static __u32 i2c_addr[] = {
	OV7720_I2CADDR,
	0x48,	/* MT9V032 default I2C address */
	0x4c,	/* MT9V032 alternate I2C address */
	0x58,	/* MT9V032 alternate I2C address */
	0x5c	/* MT9V032 alternate I2C address */
};

int check_args(int argc, char *argv[], struct grab_context *gctx)
{
	if (argc != 4)
		goto call_error;
  
	if (strcmp(argv[1],"wvga") == 0 || strcmp(argv[1],"WVGA") == 0) {
		gctx->width = 752;
		gctx->height = 480;
	}
	else if (strcmp(argv[1],"vga") == 0 || strcmp(argv[1],"VGA") == 0) {
		gctx->width = 640;
		gctx->height = 480;
	}
	else if (strcmp(argv[1],"qvga") == 0 || strcmp(argv[1],"QVGA") == 0) {
		gctx->width = 320;
		gctx->height = 240;
	}
	else
		goto call_error;
  
	if (strcmp(argv[2], GREY_STR) == 0)
		gctx->pixelformat = V4L2_PIX_FMT_GREY;
	else if (strcmp(argv[2], UYVY_STR) == 0)
		gctx->pixelformat = V4L2_PIX_FMT_UYVY;
	else if (strcmp(argv[2], YUYV_STR) == 0)
		gctx->pixelformat = V4L2_PIX_FMT_YUYV;
	else if (strcmp(argv[2], YUV422P_STR) == 0)
		gctx->pixelformat = V4L2_PIX_FMT_YUV422P;
	else if (strcmp(argv[2], GRGB_STR) == 0)
		gctx->pixelformat = V4L2_PIX_FMT_GRGB;
	else if (strcmp(argv[2], GRB422P_STR) == 0)
		gctx->pixelformat = V4L2_PIX_FMT_GRB422P;
	else if (strcmp(argv[2], SBGGR8_STR) == 0)
		gctx->pixelformat = V4L2_PIX_FMT_SBGGR8;
	else
		goto call_error;

	gctx->nr_buffers = atoi(argv[3]);
	if (gctx->nr_buffers <= 0)
		goto call_error;

	return 0;

 call_error:
	printf("Error. Call syntax is: test_grab image_format pixel_format nr_grab\n");
	printf("  image_format: one of the following strings\n");
	printf("                wvga for 752x640 image (MBS032 only),\n");
	printf("                vga  for 640x480 image,\n");
	printf("                qvga for 320x240 image (MBS7720 only).\n");
	printf("  pixel_format: one of the following strings\n");
	printf("                ""%s"" 8 bits greyscale format (MBS032M).\n",
	       GREY_STR);
	printf("                ""%s"" packed UYVY format.\n", UYVY_STR);
	printf("                ""%s"" packed YUYV format.\n", YUYV_STR);
	printf("                ""%s"" planarized YUV422 format.\n",
	       YUV422P_STR);
	printf("                ""%s"" packed GRB422 format.\n", GRGB_STR);
	printf("                ""%s"" planarized GRB422 format.\n",
	       GRB422P_STR);
	printf("                ""%s"" raw Bayer format (MBS032C).\n",
	       SBGGR8_STR);
	printf("  nr_grab     : number of grabbed images.\n");
	return -1;
}


/* set_reg is for DEBUG ONLY */
int set_reg(struct grab_context *gctx,
	    char *reg_name,
	    struct v4l2_register *reg,
	    __u64 *mask)
{
	int ret;
	__u64 new_val;

	new_val = reg->val;

	reg->match_type = V4L2_CHIP_MATCH_I2C_ADDR;
	reg->match_chip = gctx->i2c_addr;
	ret = ioctl(gctx->vfd, VIDIOC_DBG_G_REGISTER, reg);
	if (ret < 0) {
		perror("ioctl VIDIOC_DBG_G_REGISTER");
		return ret;
	}
	printf("%s (0x%x) current value = 0x%x\n",
	       reg_name, (int) reg->reg, (int) reg->val);

	reg->val &= ~(*mask);
	reg->val |= new_val;
	printf("reg->val = %x\n", (int) reg->val);
	ret = ioctl(gctx->vfd, VIDIOC_DBG_S_REGISTER, reg);
	if (ret < 0) {
		perror("ioctl VIDIOC_DBG_S_REGISTER");
		return ret;
	}
  
	usleep(500000);
	ret = ioctl(gctx->vfd, VIDIOC_DBG_G_REGISTER, reg);
	if (ret < 0) {
		perror("ioctl VIDIOC_DBG_G_REGISTER");
		return ret;
	}
	printf("%s (0x%x) NEW value = 0x%x\n",
	       reg_name, (int) reg->reg, (int) reg->val);
  
	/* wait for sensor to adjust */
	usleep(1000000);

	return 0;
}


/* disp_reg is for DEBUG ONLY */
void disp_reg(struct grab_context *gctx,
	      char *reg_name,
	      int reg_addr)
{
	int ret;
	struct v4l2_register reg;

	reg.reg = reg_addr;
	reg.match_type = V4L2_CHIP_MATCH_I2C_ADDR;
	reg.match_chip = gctx->i2c_addr;
	ret = ioctl(gctx->vfd, VIDIOC_DBG_G_REGISTER, &reg);
	if (ret < 0)
		perror("ioctl VIDIOC_DBG_G_REGISTER");
	else
		printf("%s (0x%x) current value = 0x%x\n",
		       reg_name, reg_addr, (int) reg.val);
}

void convert_GBRP_to_Bayer_and_write(struct grab_context *gctx,
				     unsigned char *gp, unsigned char *bp,
				     unsigned char *rp, FILE *f)
{
	unsigned char *g, *b, *r, *endg;
	unsigned char rgb[6];
	int jj;

	g = gp; b = bp; r = rp;
	endg = gp + gctx->width * gctx->height;

	while (g < endg) {
		/* BGBG line */
		rgb[0] = 0; rgb[1] = 0; rgb[3] = 0; rgb[5] = 0;
		for (jj=0; jj<gctx->width; jj+=2) {
			/* blue pixel */
			rgb[2] = b[jj>>1];
			/* green pixel */
			rgb[4] = g[jj+1];
			fwrite(rgb, sizeof(rgb), 1, f);
		}

		/* GRGR line */
		rgb[0] = 0; rgb[2] = 0; rgb[4] = 0; rgb[5] = 0;
		for (jj=0; jj<gctx->width; jj+=2) {
			/* green pixel */
			rgb[1] = g[jj];
			/* red pixel */
			rgb[3] = r[jj>>1];
			fwrite(rgb, sizeof(rgb), 1, f);
		}

		/* ignore next line because of data redundancy */
		g += gctx->width * 2;
		b += gctx->width;
		r += gctx->width;
	}
}

int save_image(struct grab_context *gctx)
{
	int ret;
	int ii;
	char filename[256];
	__u32 height;
	FILE *fy, *fu, *fv, *fg, *fr, *fb, *fi, *f;
	void *ptr, *ptrB, *ptrR, *end;
	unsigned char *ptr_bg, *ptr_gr, *buf_end, *line_end;
	unsigned char r, g, b, rgb[6];

	ret = 0;

	switch (gctx->pixelformat) {

	case V4L2_PIX_FMT_UYVY:
	case V4L2_PIX_FMT_YUYV:

		/* Packed YUV data format */

		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_Y_NAME);
		fy = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_U_NAME);
		fu = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_V_NAME);
		fv = fopen(filename, "wb");
		if ((fy == NULL) || (fu == NULL) || (fv == NULL)) {
			perror("fopen");
			ret = -1;
			goto save_UYVY_end;
		}
		printf("Saving images..."); fflush(stdout);
		height = gctx->height * gctx->nr_buffers;
		fprintf(fy, "P5 %d %d 255\n", gctx->width  , height);
		fprintf(fu, "P5 %d %d 255\n", gctx->width/2, height);
		fprintf(fv, "P5 %d %d 255\n", gctx->width/2, height);
		for (ii=0; ii<gctx->nr_buffers; ii++) {
			ptr = gctx->buffers[ii].start;
			end = ptr + gctx->width * gctx->height * 2;
			while (ptr < end) {
				if (gctx->pixelformat == V4L2_PIX_FMT_UYVY) {
					fwrite(ptr+0, 1, 1, fu);
					fwrite(ptr+1, 1, 1, fy);
					fwrite(ptr+2, 1, 1, fv);
					fwrite(ptr+3, 1, 1, fy);
				}
				else {
					fwrite(ptr+0, 1, 1, fy);
					fwrite(ptr+1, 1, 1, fu);
					fwrite(ptr+2, 1, 1, fy);
					fwrite(ptr+3, 1, 1, fv);
				}
				ptr += 4;
			}
		}
		printf("done.\n");
    
	save_UYVY_end:
		if (fy != NULL) fclose(fy);
		if (fu != NULL) fclose(fu);
		if (fv != NULL) fclose(fv);

		break;

	case V4L2_PIX_FMT_YUV422P:

		/* Planarized YUV422 data format:
	 
		   The planarized image buffer contains Y, U and V
		   components in different memory areas called planes.
		   For memory management reasons, each plane size is
		   an integer multiple of memory page.

		   Y plane size is width * height. It starts at the
		   base address of the image buffer which is always
		   aligned with a page boundary.

		   Next is U plane which size is (width/2) * height.
		   It starts at the page boundary following the end
		   of the Y plane.

		   Finally is V plane which size is (width/2) * height.
		   It starts at the page boundary following the end
		   of the U plane.

		   As indicated by the format name, Y is defined for
		   every pixel, U and V only for half of them: one
		   out of two on each line.

		   This format is particularly interesting for image
		   and video compression. Also, it allows separate
		   processing of luminance and chromaticities. */

		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_Y_NAME);
		fy = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_U_NAME);
		fu = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_V_NAME);
		fv = fopen(filename, "wb");
		if ((fy == NULL) || (fu == NULL) || (fv == NULL)) {
			perror("fopen");
			ret = -1;
			goto save_YUV422P_end;
		}
		printf("Saving images..."); fflush(stdout);
		height = gctx->height * gctx->nr_buffers;
		fprintf(fy, "P5 %d %d 255\n", gctx->width, height);
		fprintf(fu, "P5 %d %d 255\n", gctx->width/2, height);
		fprintf(fv, "P5 %d %d 255\n", gctx->width/2, height);
		for (ii=0; ii<gctx->nr_buffers; ii++) {
			ptr = gctx->buffers[ii].start;
			printf("Y plane start = 0x%08x\n", ptr);
			fwrite(ptr, 1, gctx->width*gctx->height, fy);
			ptr += page_align(gctx->width * gctx->height);
			printf("Cb/U plane start = 0x%08x\n", ptr);
			fwrite(ptr, 1, gctx->width*gctx->height/2, fu);
			ptr += page_align(gctx->width * gctx->height / 2);
			printf("Cr/V plane start = 0x%08x\n", ptr);
			fwrite(ptr, 1, gctx->width*gctx->height/2, fv);
		}
		printf("done.\n");

	save_YUV422P_end:
		if (fy != NULL) fclose(fy);
		if (fu != NULL) fclose(fu);
		if (fv != NULL) fclose(fv);

		break;

	case V4L2_PIX_FMT_GRB422P:

		/* Planarized GRB422 data format:
	 
		   The planarized image buffer contains G, R and B
		   components in different memory areas called planes.
		   For memory management reasons, each plane size is
		   an integer multiple of memory page.

		   G plane size is width * height. It starts at the
		   base address of the image buffer which is always
		   aligned with a page boundary.

		   Next is R plane which size is (width/2) * height.
		   It starts at the page boundary following the end
		   of the G plane.

		   Finally is B plane which size is (width/2) * height.
		   It starts at the page boundary following the end
		   of the R plane.

		   As indicated by the format name, G is defined for
		   every pixel, R and B only for half of them: one
		   out of two on each line. */

		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_G_NAME);
		fg = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_R_NAME);
		fr = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_B_NAME);
		fb = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_BAYER_NAME);
		f  = fopen(filename, "wb");
		if ((fg == NULL) || (fr == NULL) ||
		    (fb == NULL) || (f == NULL)) {
			perror("fopen");
			ret = -1;
			goto save_GRB422P_end;
		}
		printf("Saving images..."); fflush(stdout);
		height = gctx->height * gctx->nr_buffers;
		fprintf(fg, "P5 %d %d 255\n", gctx->width  , height);
		fprintf(fr, "P5 %d %d 255\n", gctx->width/2, height);
		fprintf(fb, "P5 %d %d 255\n", gctx->width/2, height);
		fprintf(f , "P6 %d %d 255\n", gctx->width  , gctx->height);
		for (ii=0; ii<gctx->nr_buffers; ii++) {
			ptr = gctx->buffers[ii].start;
			ptrR = ptr  +
				page_align(gctx->width * gctx->height);
			ptrB = ptrR +
				page_align(gctx->width * gctx->height / 2);
			printf("G plane start = 0x%08x\n", ptr);
			fwrite(ptr, 1, gctx->width*gctx->height, fg);
			printf("R plane start = 0x%08x\n", ptrR);
			fwrite(ptrR, 1, gctx->width*gctx->height/2, fr);
			printf("B plane start = 0x%08x\n", ptrB);
			fwrite(ptrB, 1, gctx->width*gctx->height/2, fb);
		}
		printf("done.\n");

		/* pseudo Bayer image */
		convert_GBRP_to_Bayer_and_write(gctx, ptr, ptrB, ptrR, f);

	save_GRB422P_end:
		if (fg != NULL) fclose(fg);
		if (fr != NULL) fclose(fr);
		if (fb != NULL) fclose(fb);
		if (f  != NULL) fclose(f );

		break;

	case V4L2_PIX_FMT_GRGB:

		/* Each line is GRGBGRGB... so 2 bytes/pixel.
		   OV7720 can out put this format. */

		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_G_NAME);
		fg = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_R_NAME);
		fr = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_B_NAME);
		fb = fopen(filename, "wb");
		if ((fg == NULL) || (fr == NULL) || (fb == NULL)) {
			perror("fopen");
			ret = -1;
			goto save_GRGB_end;
		}
		printf("Saving images..."); fflush(stdout);
		height = gctx->height * gctx->nr_buffers;
		fprintf(fg, "P5 %d %d 255\n", gctx->width  , height);
		fprintf(fr, "P5 %d %d 255\n", gctx->width/2, height);
		fprintf(fb, "P5 %d %d 255\n", gctx->width/2, height);
		for (ii=0; ii<gctx->nr_buffers; ii++) {
			ptr = gctx->buffers[ii].start;
			end = ptr + gctx->width * gctx->height * 2;
			while (ptr < end) {
				fwrite(ptr+0, 1, 1, fg);
				fwrite(ptr+1, 1, 1, fr);
				fwrite(ptr+2, 1, 1, fg);
				fwrite(ptr+3, 1, 1, fb);
				ptr += 4;
			}
		}
		printf("done.\n");

	save_GRGB_end:
		if (fg != NULL) fclose(fg);
		if (fr != NULL) fclose(fr);
		if (fb != NULL) fclose(fb);

		break;

	case V4L2_PIX_FMT_GREY:
#ifdef SAVE_SBGGR8_AS_Y8
	case V4L2_PIX_FMT_SBGGR8:
#endif
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_Y_NAME);
		fy = fopen(filename, "wb");
		if (fy == NULL) {
			perror("fopen");
			ret = -1;
			goto save_GREY_end;
		}
		printf("Saving images..."); fflush(stdout);
		height = gctx->height * gctx->nr_buffers;
		fprintf(fy, "P5 %d %d 255\n", gctx->width, height);
		for (ii=0; ii<gctx->nr_buffers; ii++) {
			ptr = gctx->buffers[ii].start;
			printf("Y plane start = 0x%08x\n", ptr);
			fwrite(ptr, 1, gctx->width*gctx->height, fy);
		}
		printf("done.\n");

	save_GREY_end:
		if (fy != NULL) fclose(fy);

		break;

#ifndef SAVE_SBGGR8_AS_Y8
	case V4L2_PIX_FMT_SBGGR8:

		/* Bayer data format (output format of MBS032C):
		   first line GRGR...
		   second line BGBG...
		*/

		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_G_NAME);
		fg = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_R_NAME);
		fr = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_B_NAME);
		fb = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_NAME);
		fi  = fopen(filename, "wb");
		sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_BAYER_NAME);
		f  = fopen(filename, "wb");
		if ((fg == NULL) || (fr == NULL) || (fb == NULL) || (fi == NULL) ||
		    (f == NULL)) {
			perror("fopen");
			ret = -1;
			goto save_SBGGR8_end;
		}
		printf("Saving images..."); fflush(stdout);
		height = gctx->height * gctx->nr_buffers;
		fprintf(fg, "P5 %d %d 255\n", gctx->width/2, height);
		fprintf(fr, "P5 %d %d 255\n", gctx->width/2, height/2);
		fprintf(fb, "P5 %d %d 255\n", gctx->width/2, height/2);
		fprintf(f , "P6 %d %d 255\n", gctx->width  , height);
		fprintf(fi, "P6 %d %d 255\n", gctx->width  , height);

		for (ii=0; ii<gctx->nr_buffers; ii++) {
			ptr_gr = (unsigned char*) gctx->buffers[ii].start;
			buf_end = ptr_gr + gctx->width * gctx->height;
			while (ptr_gr < buf_end) {
				/* GRGR... line */
				ptr_bg = ptr_gr + gctx->width;
				line_end = ptr_bg;
				while (ptr_gr < line_end) {
					r = *(ptr_gr+1);
					g = *(ptr_gr+0);
					b = *(ptr_bg+0);
					fwrite(&g, 1, 1, fg);
					fwrite(&r, 1, 1, fr);
					rgb[0] = rgb[3] = r;
					rgb[1] = rgb[4] = g;
					rgb[2] = rgb[5] = b;
					fwrite(rgb, sizeof(rgb), 1, fi);
					rgb[0] = rgb[2] = 0;
					rgb[1] = g;
					rgb[3] = r;
					rgb[4] = rgb[5] = 0;
					fwrite(rgb, sizeof(rgb), 1, f);
					ptr_bg += 2;
					ptr_gr += 2;
				}
				/* BGBG... line */
				ptr_gr = line_end - gctx->width;
				ptr_bg = line_end;
				while (ptr_gr < line_end) {
					r = *(ptr_gr+1);
					g = *(ptr_bg+1);
					b = *(ptr_bg+0);
					fwrite(&b, 1, 1, fb);
					fwrite(&g, 1, 1, fg);
					rgb[0] = rgb[3] = r;
					rgb[1] = rgb[4] = g;
					rgb[2] = rgb[5] = b;
					fwrite(rgb, sizeof(rgb), 1, fi);
					rgb[0] = rgb[1] = 0;
					rgb[2] = b;
					rgb[3] = rgb[5] = 0;
					rgb[4] = g;
					fwrite(rgb, sizeof(rgb), 1, f);
					ptr_bg += 2;
					ptr_gr += 2;
				}
				ptr_gr += gctx->width;
			}
		}
		printf("done.\n");

	save_SBGGR8_end:
		if (fg != NULL) fclose(fg);
		if (fr != NULL) fclose(fr);
		if (fb != NULL) fclose(fb);
		if (fi != NULL) fclose(fi);
		if (f  != NULL) fclose(f );

	break;
#endif
	default:
		printf("??? %d unknown pixel format.\n", gctx->pixelformat);
		ret = -1;
	}
 	  
	return ret;
}

int grab_start(struct grab_context *gctx)
{
	int ret;
	int ii;
	struct v4l2_buffer buffer;

	memset(&buffer, 0, sizeof (buffer));
	buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buffer.memory = V4L2_MEMORY_MMAP;

	for (ii=0; ii<gctx->nr_buffers; ii++) {
		buffer.index = ii;
		ret = ioctl(gctx->vfd, VIDIOC_QBUF, &buffer);
		if (ret < 0) {
			perror ("ioctl VIDIOC_QBUF");
			return ret;
		}

		printf("buffer %u: bytesused = %u\n",
		       buffer.index, buffer.bytesused);
	}

	ret = ioctl(gctx->vfd, VIDIOC_STREAMON, &buffer.type);
	if (ret < 0) {
		perror ("ioctl VIDIOC_STREAMON");
		return ret;
	}
  
	printf("Wait for sensor auto adjust...");
	fflush(stdout);
	sleep(1);
	printf(" Done.\n");

	return 0;
}

int grab_stop(struct grab_context *gctx)
{
	int ret;
	enum v4l2_buf_type type;

	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	ret = ioctl(gctx->vfd, VIDIOC_STREAMOFF, &type);
	if (ret < 0) {
		perror ("ioctl VIDIOC_STREAMOFF");
		return ret;
	}

	return 0;
}

int grab_image(struct grab_context *gctx)
{
	int ret;
	int ii;
	struct v4l2_buffer buffer;
	struct timeval tvs, tve, dt, tv;
  
	/* start capture */
	ret = grab_start(gctx);
	if (ret < 0)
		return ret;

	buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buffer.memory = V4L2_MEMORY_MMAP;

	for (ii=0; ii<gctx->nr_buffers; ii++) {
		/* retrieve image when sensor turned on */
		buffer.index = ii;
		ret = ioctl(gctx->vfd, VIDIOC_DQBUF, &buffer);
		if (ret < 0) {
			if (errno == EAGAIN)
				printf("Image not grabbed yet.\n");
			else
				perror("ioctl VIDIOC_DQBUF");
			return ret;
		}
	}

	/* queue all buffers first */
	int jj;
	for (jj=0; jj<1; jj++) {
		for (ii=0; ii<gctx->nr_buffers; ii++) {
			buffer.index = ii;
			gettimeofday(&tvs, NULL);
			ret = ioctl(gctx->vfd, VIDIOC_QBUF, &buffer);
			if (ret < 0) {
				perror ("ioctl VIDIOC_QBUF");
				goto stream_off;
			}
			gettimeofday(&tve, NULL);
			timersub(&tve, &tvs, &dt);
			printf("QBUF duration = %ld.%06d\n",
			       dt.tv_sec, dt.tv_usec);
			printf("buffer %u: bytesused = %u\n",
			       buffer.index, buffer.bytesused);
			usleep(5000);
		}

		/* dequeue buffers */
		for (ii=0; ii<gctx->nr_buffers; ii++) {
			buffer.index = ii;
			ret = ioctl(gctx->vfd, VIDIOC_DQBUF, &buffer);
			if (ret < 0) {
				if (errno == EAGAIN)
					printf("Image not grabbed yet.\n");
				else
					perror("ioctl VIDIOC_DQBUF");
				goto stream_off;
			}
			printf("%ld.%06ld DQBUF %02d end, frame %u.\n",
			       buffer.timestamp.tv_sec,
			       buffer.timestamp.tv_usec,
			       buffer.index, buffer.sequence);
		}

		if (jj < gctx->nr_buffers-1) {
			printf("Waiting 1s before next acquisition\n");
			usleep(1000000);
		}
	}

	ret = save_image(gctx);
	if (ret < 0)
		goto stream_off;

 stream_off:
	/* stop capture */
	ret = grab_stop(gctx);
	if (ret < 0)
		return ret;

	return 0;
}


int grab_init(struct grab_context *gctx)
{
	int ret;
	int input;
	int ii, nb_i2c_addr;
	struct v4l2_register reg;
	__u64 mask;
	struct v4l2_chip_ident chip_id;
	struct v4l2_control ctrl;
	struct v4l2_format format;
	struct v4l2_requestbuffers reqbuf;
	struct v4l2_buffer buffer;

	gctx->vfd = -1;
	gctx->buffers = NULL;

	gctx->vfd = open(VIDEO_DEV_NAME, O_RDWR);
	if (gctx->vfd < 0) {
		printf("%s ", VIDEO_DEV_NAME);
		perror("open failed");
		return -1;
	}

	printf("%s open succeeded.\n", VIDEO_DEV_NAME);

	/* select input 0, the only one */
	input = 0;

	ret = ioctl(gctx->vfd, VIDIOC_S_INPUT, &input);
	if (ret != 0) {
		perror("ioctl VIDIOC_S_INPUT");
		return ret;
	}

	/* detect sensor chip ID */
	nb_i2c_addr = sizeof(i2c_addr) / sizeof(i2c_addr[0]);
	gctx->i2c_addr = -1;
	chip_id.match_type = V4L2_CHIP_MATCH_I2C_ADDR;
	for (ii=0; ii<nb_i2c_addr; ii++) {
		chip_id.match_chip = i2c_addr[ii];
		ret = ioctl(gctx->vfd, VIDIOC_G_CHIP_IDENT, &chip_id);
		if (ret == 0) {
			gctx->i2c_addr = chip_id.match_chip;
			break;
		}
	}
	if (gctx->i2c_addr == -1) {
		printf("No sensor found. Make sure camera is connected.\n");
		return -1;
	}

	/* specific settings (test oonly) */
	if (gctx->i2c_addr == OV7720_I2CADDR) {
#if 0
		/* Clock prescaler: CLKRC
		   CLKRC<6> = 1: no prescale, PCLK = MCLK
		   CLKRC<6> = 0: prescale defined by PCLK = 2*MCLK/(CLKRC<5:0>+1)
		   examples:
		   CLKRC<6:0> = 0x00 => fastest rate: PCLK = 2*MCLK
		   CLKRC<6:0> = 0x01 => PCLK = MCLK
		   CLKRC<6:0> = 0x02 => PCLK = MCLK/2
		   CLKRC<6:0> = 0x03 => PCLK = MCLK*2/3
		   CLKRC<6:0> = 0x3f => slowest rate: PCLK = MCLK/16 */
		reg.reg = OV7720_CLKRC;
		mask = 0x8f; /* bits <6:0> */
		reg.val = 0x01;
		set_reg(gctx, "CLKRC", &reg, &mask);
#endif
	}
	else {
#if 0
		/* MT9V022 test */
		reg.reg = 0x70;
		reg.val = 0;
		mask = 0x20;
		set_reg(gctx, "Row Noise Correction Ctrl 1", &reg, &mask);
		reg.reg = 0x7f;
		reg.val = 0x2400;	/* black image */
		reg.val = 0x27ff;	/* white image */
		reg.val = 0x2600;	/* grey 128 image */
		reg.val = 0x67ff;	/* white/black image */
		reg.val = 0x6500;	/* grey64/grey191 image */
		reg.val = 0x2800;	/* vert */
		reg.val = 0x3000;	/* horiz */
		reg.val = 0x3800;	/* diag */
		reg.val = 0x6404;
		mask = 0xffff;
		set_reg(gctx, "Digital Test Pattern", &reg, &mask);
#endif
	}

	/* configure acquisition format */
	format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	format.fmt.pix.width = gctx->width;
	format.fmt.pix.height = gctx->height;
	format.fmt.pix.pixelformat = gctx->pixelformat;
	format.fmt.pix.field = V4L2_FIELD_NONE;

	ret = ioctl(gctx->vfd, VIDIOC_S_FMT, &format);
	if (ret != 0) {
		perror("ioctl VIDIOC_S_FMT");
		return ret;
	}

	/* request buffers to store grabbed image */
	reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	reqbuf.memory = V4L2_MEMORY_MMAP;
	reqbuf.count = gctx->nr_buffers;

	ret = ioctl(gctx->vfd, VIDIOC_REQBUFS, &reqbuf);
	if (ret < 0) {
		if (errno == EINVAL)
			printf ("Video capturing or mmap-streaming is not supported\n");
		else
			perror ("ioctl VIDIOC_REQBUFS");
		return ret;
	}

	if (reqbuf.count < gctx->nr_buffers) {
		printf("   !!! Only %d buffers available\n", reqbuf.count);
		gctx->nr_buffers = reqbuf.count;
	}

	gctx->buffers = calloc(gctx->nr_buffers, sizeof (*gctx->buffers));
	if (gctx->buffers == NULL) {
		perror("calloc gctx->buffers");
		return -1;
	}

	for (ii=0; ii<gctx->nr_buffers; ii++) {
		/* query buffer to allocate and mmap it */
		memset(&buffer, 0, sizeof (buffer));
		buffer.type = reqbuf.type;
		buffer.memory = V4L2_MEMORY_MMAP;
		buffer.index = ii;

		ret = ioctl(gctx->vfd, VIDIOC_QUERYBUF, &buffer);
		if (ret < 0) {
			perror ("ioctl VIDIOC_QUERYBUF");
			exit (EXIT_FAILURE);
		}

		printf("buffer %u: length = %u, bytesused = %u\n",
		       buffer.index, buffer.length, buffer.bytesused);

		/* remember for munmap */
		gctx->buffers[ii].length = buffer.length;
		gctx->buffers[ii].start = mmap (NULL, buffer.length,
						PROT_READ | PROT_WRITE,
						MAP_SHARED,
						gctx->vfd, buffer.m.offset);
		if (MAP_FAILED == gctx->buffers[ii].start) {
			perror ("mmap");
			return -1;
		}
	}

	return 0;
}

void grab_end(struct grab_context *gctx)
{
	int ii;
	if (gctx->buffers != NULL) {
		for (ii=0; ii<gctx->nr_buffers; ii++) {
			if ((gctx->buffers[ii].start != MAP_FAILED) &&
			    (gctx->buffers[ii].length > 0))
				munmap (gctx->buffers[ii].start,
					gctx->buffers[ii].length);
		}
    
		free(gctx->buffers);
	}

	if (gctx->vfd != -1) close(gctx->vfd);
}

int main(int argc, char *argv[])
{
	int ret;
	struct grab_context gctx;
	struct v4l2_input input;
  
	printf("\n\n   ***   test_grab   ***\n\n");
	printf("Basic image grab with V4L2 compatible PXA camera driver\n\n");
	usleep(100000); /* for proper printing */

	ret = check_args(argc, argv, &gctx);
	if (ret) exit(1);

	ret = grab_init(&gctx);
	if (ret < 0) goto end;

	ret = grab_image(&gctx);
	if (ret < 0) goto end;
	/*   frame_drop(&gctx); */

 end:
	grab_end(&gctx);

	exit(0);
}
