#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/gpiodev.h>

#define GPIO_DEV_NAME	"/dev/gpio"

#define CMD_SET_GPIO	1
#define CMD_CLEAR_GPIO	2
#define CMD_GET_GPIO	3

struct mg_context {
	int fd;
	int gpio;
	int state;
	int cmd;
	int duration;
};


int mg_load_args(int argc, char *argv[], struct mg_context *ctx)
{
	int ii;

	if (argc < 3)
		goto call_error;

	ctx->gpio = atoi(argv[1]);
	if (ctx->gpio < 0)
		goto call_error;

	if ((strcmp(argv[2], "SET") == 0) || (strcmp(argv[2], "set") == 0))
		ctx->cmd = CMD_SET_GPIO;
	else if ((strcmp(argv[2], "CLEAR") == 0) || (strcmp(argv[2], "clear") == 0))
		ctx->cmd = CMD_CLEAR_GPIO;
	else if ((strcmp(argv[2], "GET") == 0) || (strcmp(argv[2], "get") == 0))
		ctx->cmd = CMD_GET_GPIO;
	else
		goto call_error;

	ctx->duration = -1;
	if (argc > 3) {
		if (isdigit(argv[3][0])) {
			ctx->duration = atoi(argv[3]);
			if (ctx->duration <= 0)
				goto call_error;
		}
		else
			goto call_error;
	}

	return 0;

 call_error:
	printf("mbs-gpio: a command to access MBS270 GPIOs.\n");
	printf("Syntax:\n");
	printf("  mbs-gpio gpio set [ms]    Sets the corresponding GPIO output\n");
	printf("                            for ms milliseconds optionnally\n");
	printf("  mbs-gpio gpio clear [ms]  Clears the corresponding GPIO output\n");
	printf("                            for ms milliseconds optionnally\n");
	printf("  mbs-gpio gpio get         Gets the corresponding GPIO input state\n");
	return -1;
}

int mg_init(struct mg_context *ctx)
{
	struct gpiodev_add gpio_add;

	ctx->state = -1;
  
	ctx->fd = open(GPIO_DEV_NAME, O_RDWR);
	if (ctx->fd < 0) {
		char err_str[64];
		sprintf(err_str, "open %s", GPIO_DEV_NAME);
		perror(err_str);
		return -1;
	}

	gpio_add.gpio = ctx->gpio;
	if ((ctx->cmd == CMD_SET_GPIO) || (ctx->cmd == CMD_CLEAR_GPIO)) {
		gpio_add.input = 0;
		gpio_add.init_state = (ctx->cmd == CMD_SET_GPIO);
		gpio_add.permanent = (ctx->duration < 0);
		if (ioctl(ctx->fd, IOCGPIODEVADD, (void*) &gpio_add)) {
			perror("ioctl IOCGPIODEVADD: ");
			return -1;
		}
		ctx->state = 0;
	}
	else if (ctx->cmd == CMD_GET_GPIO) {
		gpio_add.input = 1;
		if (ioctl(ctx->fd, IOCGPIODEVADD, (void*) &gpio_add)) {
			perror("ioctl IOCGPIODEVADD: ");
			return -1;
		}
		ctx->state = (gpio_add.init_state != 0);
	}
	else
		return -1;

	return 0;
}

int mg_do(struct mg_context *ctx)
{
	if (ctx->duration > 0) {
		usleep(ctx->duration * 1000);
		if (ctx->cmd == CMD_SET_GPIO)
			ioctl(ctx->fd, IOCGPIODEVCLR, (void*) &ctx->gpio);
		else if (ctx->cmd == CMD_CLEAR_GPIO)
			ioctl(ctx->fd, IOCGPIODEVSET, (void*) &ctx->gpio);
	}

	return 0;
}

int mg_end(struct mg_context *ctx)
{
	int ret;

	ret = 0;

	if (ctx->state >= 0) {
		if (ioctl(ctx->fd, IOCGPIODEVREM, (void*) &ctx->gpio)) {
			perror("ioctl IOCGPIODEVREM: ");
			ret = -1;
		}
	}
	if (ctx->fd >= 0)
		close(ctx->fd);

	return ret;
}

int main(int argc, char *argv[])
{
	int exit_code;
	struct mg_context ctx;

	if (mg_load_args(argc, argv, &ctx))
		exit(-1);

	exit_code = 0;
	if (mg_init(&ctx))
		exit_code = -1;
	else
		mg_do(&ctx);

	if (mg_end(&ctx))
		exit_code = -1;

	if (exit_code)
		exit(exit_code);
	else
		exit(ctx.state);
}
