#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>

#define M41T0_REG_SECONDS	8
#define M41T0_REG_MINUTES	1
#define M41T0_REG_CENT_HOURS	2
#define M41T0_REG_WEEKDAY	3
#define M41T0_REG_DAY		4
#define M41T0_REG_MONTH		5
#define M41T0_REG_YEAR		6
#define M41T0_REG_CTRL		7

#define I2C_DEV_NAME "/dev/i2c-"

typedef struct im_context {
	int		fd;
	int		i2c_port;
	unsigned int	i2c_addr;
} im_context_t;

int im_check_args(im_context_t *ctx, int argc, char *argv[])
{
	if (argc != 3)	
		goto call_error;

	ctx->i2c_port = (unsigned int) atoi(argv[1]);
	if (ctx->i2c_port < 0 || ctx->i2c_port > 255)
		goto call_error;

	sscanf(argv[2], "%x", &ctx->i2c_addr);
	if (ctx->i2c_addr < 0 || ctx->i2c_addr > 0xff)
		goto call_error;

	return 0;

 call_error:
	printf("Error. Call syntax is: init_m41t0 i2c_port i2c_addr\n");
	printf("   I2C port (value between 0 and 255 after %s)\n", I2C_DEV_NAME);
	printf("   M41T0 I2C device address (0x68)\n");
	return -1;
}

int im_init(im_context_t *ctx)
{
	char dev_name[sizeof(I2C_DEV_NAME)+4];

	sprintf(dev_name, "%s%d", I2C_DEV_NAME, ctx->i2c_port);
	ctx->fd = open(dev_name, O_RDWR);
	if (ctx->fd < 0) {
		printf("%s ", dev_name);
		perror("open failed");
		return -1;
	}

	/* select device */
	if (ioctl(ctx->fd, I2C_SLAVE, (int) ctx->i2c_addr) < 0) {
		printf("select device 0x%x\n", ctx->i2c_addr);
		perror("ioctl I2C_SLAVE");
		return -1;
	}

	printf("Connected to device 0x%x on %s.\n", ctx->i2c_addr, dev_name);
	return 0;
}

int im_do(im_context_t *ctx)
{
	char str[3];
	time_t current_time;
	struct tm tm_time;
	struct { char reg, val;} buf;

	/* stop M41T0 oscillator */
	buf.reg = M41T0_REG_SECONDS;
	buf.val = 0x80;
	if (write(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("write seconds register");
		return -1;
	}

	current_time = time(NULL);
	gmtime_r(&current_time, &tm_time);

	printf("%d/%d/%d %d:%d:%d wd=%d\n",
	       tm_time.tm_mday, tm_time.tm_mon, tm_time.tm_year,
	       tm_time.tm_hour, tm_time.tm_min, tm_time.tm_sec,
	       tm_time.tm_wday);

	/* year */
	buf.reg = M41T0_REG_YEAR;
	sprintf(str, "%02d", tm_time.tm_year - 100);
	printf("year: %s\n", str);
	buf.val = ((str[0] - 0x30) << 4) | (str[1] - 0x30);
	if (write(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("write year register");
		return -1;
	}

	/* month */
	buf.reg = M41T0_REG_MONTH;
	sprintf(str, "%02d", tm_time.tm_mon + 1);
	printf("month: %s\n", str);
	buf.val = ((str[0] - 0x30) << 4) | (str[1] - 0x30);
	if (write(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("write month register");
		return -1;
	}

	/* day */
	buf.reg = M41T0_REG_DAY;
	sprintf(str, "%02d", tm_time.tm_mday);
	printf("day: %s\n", str);
	buf.val = ((str[0] - 0x30) << 4) | (str[1] - 0x30);
	if (write(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("write day register");
		return -1;
	}

	/* weekday */
	buf.reg = M41T0_REG_WEEKDAY;
	sprintf(str, "%d", tm_time.tm_wday + 1);
	printf("weekday: %s\n", str);
	buf.val = str[0] - 0x30;
	if (write(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("write weekday register");
		return -1;
	}

	/* century - hours */
	buf.reg = M41T0_REG_CENT_HOURS;
	sprintf(str, "%02d", tm_time.tm_hour);
	printf("hours: %s\n", str);
	buf.val = ((str[0] - 0x30) << 4) | (str[1] - 0x30);
	if (write(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("write hours register");
		return -1;
	}

	/* minutes */
	buf.reg = M41T0_REG_MINUTES;
	sprintf(str, "%02d", tm_time.tm_min);
	printf("minutes: %s\n", str);
	buf.val = ((str[0] - 0x30) << 4) | (str[1] - 0x30);
	if (write(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("write minutes register");
		return -1;
	}

	/* seconds - starts also oscillator (clears ST bit) */
	buf.reg = M41T0_REG_SECONDS;
	sprintf(str, "%02d", tm_time.tm_sec);
	printf("seconds: %s\n", str);
	buf.val = ((str[0] - 0x30) << 4) | (str[1] - 0x30);
	if (write(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("write seconds register");
		return -1;
	}

	/* wait at least 4s before clearing the oscillator fail (OF) bit */
	int delay_sec = 10;
	printf("Waiting %d seconds before clearing OF bit...\n", delay_sec);
	sleep(delay_sec);
	printf("Done. Clearing OF bit...\n");
	buf.reg = M41T0_REG_MINUTES;
	if (write(ctx->fd, (void*) &buf, 1) != 1) {
		perror("write for read minutes register");
		return -1;
	}
	if (read(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("read minutes register");
		return -1;
	}
	buf.val &= ~0x80;
	if (write(ctx->fd, (void*) &buf, sizeof(buf)) != sizeof(buf)) {
		perror("write seconds register with OF bit clear");
		return -1;
	}
	printf("Done.\n");

	return 0;
}

void im_stop(im_context_t *ctx)
{
	if (ctx->fd >= 0)
		close(ctx->fd);
}

int main(int argc, char *argv[])
{
	im_context_t ctx;
	unsigned int addr, reg, val;
	char buf[2], op[2];

	printf("\n\n   ***   init_m41t0   ***\n\n");
	printf("This utility program initializes M41T0 RTC with current system date.\n\n");

	if (im_check_args(&ctx, argc, argv))
		exit(-1);

	if (im_init(&ctx))
		goto im_end;

	im_do(&ctx);

 im_end:
	im_stop(&ctx);

	exit(0);
}
